from PyQt5 import QtWidgets, QtCore
from .helper import logger
from .core import Nav


class NavSettings(QtWidgets.QDialog):
    settings_changed = QtCore.pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        preflist = QtWidgets.QListWidget()
        self.stack = QtWidgets.QStackedWidget()
        self.set_items = ["General", "Panes",  # Tabs", "Sorting", "Renaming",
                          # "Colors",
                          ]
        for item in self.set_items:
            preflist.addItem(item)
            self.set_ui(item)

        applyButton = QtWidgets.QPushButton('Apply')
        applyButton.clicked.connect(self.apply)
        okButton = QtWidgets.QPushButton('OK')
        okButton.clicked.connect(self.ok)
        cancelButton = QtWidgets.QPushButton('Cancel')
        cancelButton.clicked.connect(self.hide)

        mainLayout = QtWidgets.QGridLayout(self)
        mainLayout.addWidget(preflist, 0, 0, QtCore.Qt.AlignLeft)
        mainLayout.addWidget(self.stack, 0, 1, 1, 4, QtCore.Qt.AlignTop)
        mainLayout.addWidget(applyButton, 1, 4)
        mainLayout.addWidget(okButton, 1, 5)
        mainLayout.addWidget(cancelButton, 1, 6)
        self.setWindowTitle('Settings')
        self.setGeometry(300, 300, 800, 200)
        preflist.currentRowChanged.connect(self.display)
        self.sizeHint()

    def set_ui(self, item):
        layout = QtWidgets.QHBoxLayout()
        stacks = QtWidgets.QWidget()
        if item == "General":
            folders_first = QtWidgets.QCheckBox("Sort folders first")
            if Nav.conf["sort_folders_first"]:
                folders_first.setChecked(True)
            layout.addWidget(folders_first)
        elif item == "Panes":
            for i in range(1, 5):
                name = f"Pane {i}"
                pane_cb = QtWidgets.QCheckBox(name)
                if Nav.conf["panes"][name]["visible"]:
                    pane_cb.setChecked(True)
                layout.addWidget(pane_cb)
        stacks.setLayout(layout)
        self.stack.addWidget(stacks)

    def apply(self):
        for c in self.stack.children():
            for widget in c.children():
                if isinstance(widget, QtWidgets.QCheckBox):
                    logger.debug(f"CB: {widget.text()}-{widget.checkState()}")
                    if widget.text() == "Sort folder to top":
                        if widget.checkState():
                            Nav.conf["sort_folders_first"] = True
                        else:
                            Nav.conf["sort_folders_first"] = False
                        logger.debug(Nav.conf["sort_folders_first"])

                    if widget.text() in ["Pane 1", "Pane 2", "Pane 3",
                                         "Pane 4"]:
                        if widget.checkState():
                            Nav.conf["panes"][widget.text()]["visible"] = True
                        else:
                            Nav.conf["panes"][widget.text()]["visible"] = False
        self.settings_changed.emit()

    def ok(self):
        self.apply()
        self.hide()

    def display(self, i):
        self.stack.setCurrentIndex(i)



import os
import pathlib
from .helper import logger
from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler


class NavWatcher:
    """Helper class to start/stop watchdog and add/remove paths."""
    running = False
    monitored = {}
    event_handler = FileSystemEventHandler()
    observer = Observer()

    @classmethod
    def on_file_system_event(cls, event):
        """Invokes provided callback on FileSystemEvent."""
        loc = str(pathlib.PurePath(event.src_path).parent)
        try:
            for callback in cls.monitored[loc]['callbacks']:
                callback(event, loc)
        except KeyError:
            # KeyError occurs for parent folders. Expected
            # logger.warning(f"No callbacks available for {loc}")
            pass

    @classmethod
    def add_path(cls, loc, callback):
        """Adds callbacks and watches for the said location."""
        if not os.path.exists(loc):
            logger.warning(f"{loc} no longer exists.")
            return
        try:
            if callback not in cls.monitored[loc]['callbacks']:
                cls.monitored[loc]['callbacks'].append(callback)
            logger.debug(f"Callback {callback} registered for {loc}")
        except KeyError:
            cls.monitored[loc] = {
                "stamp": os.stat(loc).st_mtime,
                "callbacks": [callback],
                "watch": cls.observer.schedule(cls.event_handler, loc),
            }
            logger.debug(f"Monitoring started for {loc}")
        logger.debug(f"Current watchers: {cls.observer._watches}")

    @classmethod
    def remove_path(cls, loc, callback):
        """Removes callbacks and watches for the said location."""
        try:
            cls.monitored[loc]['callbacks'].remove(callback)
            logger.debug(f"Callback {callback} removed from {loc}")
        except (KeyError, ValueError):
            logger.warning(f"Error removing callback for {loc}")
        try:
            if "callbacks" not in cls.monitored[loc] or \
                    len(cls.monitored[loc]["callbacks"]) < 1:
                cls.observer.unschedule(cls.monitored[loc]["watch"])
                logger.debug(f"Stopped monitoring for {loc} as no callbacks")
                del cls.monitored[loc]
        except (KeyError, ValueError):
            logger.warning(f"Error unscheduling watch for {loc}.")
        logger.debug(f"Current watchers: {cls.observer._watches}")

    @classmethod
    def start(cls):
        """Starts the watchdog and hooks function for various events."""
        if cls.running is False:
            cls.running = True
            cls.event_handler.on_deleted = cls.on_file_system_event
            cls.event_handler.on_moved = cls.on_file_system_event
            cls.event_handler.on_created = cls.on_file_system_event
            cls.event_handler.on_modified = cls.on_file_system_event
            cls.observer.start()

    @classmethod
    def stop(cls):
        """Unschedules all watchdogs."""
        cls.observer.unschedule_all()
        cls.running = False



import os
import pathlib
from PyQt5 import QtWidgets, QtCore
from .core import Nav
from .custom import NavTree
from .helper import logger
from .navwatcher import NavWatcher
from .pub import Pub
from .tabs import NavTabWidget


class NavPane(QtWidgets.QFrame):
    pane_updated = QtCore.pyqtSignal(QtCore.QObject)

    def __init__(self, pid, pane_info):
        super().__init__()
        x = self.size().width()
        self.pid = pid
        self.location = str(pathlib.Path.home())
        self.setStyleSheet("border: 1px solid red;")
        self.setStyleSheet(
                    "NavPane {border: 1px solid ;}")
        grid = QtWidgets.QGridLayout(self)
        self.abar = QtWidgets.QLineEdit()
        self.sb = QtWidgets.QStatusBar()
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.tree = NavTree()
        self.tree.clicked[QtCore.QModelIndex].connect(self.tree_navigate)
        self.abar.returnPressed.connect(self.go_to)
        self.tabbar = NavTabWidget(self)
        # Create button that must be placed in tabs row
        new_tab = QtWidgets.QToolButton()
        new_tab.setText("+")
        self.tabbar.setCornerWidget(new_tab, QtCore.Qt.TopRightCorner)
        new_tab.clicked.connect(lambda: self.tabbar.new_tab())
        try:
            logger.debug(f"{self.pid}: Restoring tabs")
            for i in range(pane_info["tabs"]["total"]):
                tab_info = pane_info["tabs"][str(i)]
                self.tabbar.new_tab(tab_info)
        except KeyError:
            logger.error(f"{self.pid}: Error restoring tabs")
            self.tabbar.new_tab()
        try:
            logger.debug(f"Selecting active tab")
            self.tabbar.setCurrentIndex(int(pane_info["tabs"]["active"]))
        except KeyError:
            logger.error(f"Error Selecting active tab", exc_info=True)

        # line edit for filtering
        self.filter_edit = QtWidgets.QLineEdit()
        self.filter_edit.textChanged.connect(
            lambda: self.tabbar.currentWidget().set_filter(
                    self.filter_edit.text()))
        self.splitter.addWidget(self.tree)
        self.splitter.addWidget(self.tabbar)
        self.splitter.setSizes(Nav.getsizes(self.pid, [20, 80],
                                            [x*0.2, x*0.8]))
        grid.addWidget(self.abar, 0, 0, 1, 2)
        grid.addWidget(self.splitter, 1, 0, 1, 2)
        grid.addWidget(self.filter_edit, 2, 0, 1, 1)
        grid.addWidget(self.sb, 2, 1, 1, 1)
        self.tabbar.currentChanged.connect(self.tab_changed)
        self.setLayout(grid)
        if pane_info["visible"]:
            Pub.subscribe(f"Panes.{self.pid}", self.update_status_bar)
            self.update_gui(self.tabbar.currentWidget().location)
            # start monitoring active tab for refreshing
            NavWatcher.add_path(self.location, self.change_detected)
            NavWatcher.start()
        self.installEventFilter(self)  # this will catch focus events
        self.tabbar.tab_created.connect(lambda: self.install_filters())

    def set_visibility(self, visibility):
        """Toggle pane visibility."""
        if visibility:
            self.location = self.tabbar.currentWidget().location
            self.update_gui(self.location)
            Pub.subscribe(f"Panes.{self.pid}", self.update_status_bar)
            NavWatcher.add_path(self.location, self.change_detected)
            self.tabbar.currentWidget().load_tab()
        else:
            NavWatcher.remove_path(self.location, self.change_detected)
            Pub.unsubscribe(f"Panes.{self.pid}", self.update_status_bar)

    def change_detected(self, evt, loc: str):
        """Informs current tab if its current directory was changed."""
        if loc == self.location:
            self.tabbar.currentWidget().change_detected(evt)
            self.sb.showMessage(self.tabbar.currentWidget().status_info)

    def eventFilter(self, obj, event):
        """Reimplemented to handle active pane."""
        if event.type() == QtCore.QEvent.MouseButtonPress:
            # logger.debug(f"{self.pid}: {event} {event.type()} for {obj}")
            self.location = self.tabbar.currentWidget().location
            try:
                if os.getcwd() != self.location:
                    os.chdir(self.location)
            except FileNotFoundError:
                pass
            except NotADirectoryError:
                logger.error(f"{self.location} is not a directory")
            self.pane_updated.emit(self)
            return QtCore.QObject.eventFilter(self, obj, event)
        else:
            return False

    def showEvent(self, event):
        """Installs event filters in all children of panel."""
        super().showEvent(event)
        # this will install event filter in all children of the panel
        self.install_filters()

    def install_filters(self):
        """Install event filter in all children of the panel."""
        for widget in self.findChildren(QtWidgets.QWidget):
            widget.installEventFilter(self)

    def tree_navigate(self, index):
        """Handles tree navigation."""
        loc = self.tree.model.filePath(index)
        logger.debug(f"{self.pid}: Tree Navigation: {loc}")
        ret = self.tabbar.currentWidget().navigate(loc)
        self.sb.showMessage(self.tabbar.currentWidget().status_info)
        if ret:
            self.abar.setText(loc)
            self.location = loc

    def go_to(self):
        """Handles address bar navigation."""
        loc = self.abar.text().replace('\\', '/')
        loc = loc.strip()
        if loc.startswith("@"):
            if "=" in loc:  # alias is being set
                alias, actual = loc.split("=", maxsplit=1)
                alias = alias.strip()
                actual = actual.strip()
                if actual == "":
                    try:
                        del Nav.conf["aliases"][alias]
                        logger.debug(f"Alias: {alias} unset")
                        Pub.notify("App", f"{self.pid}: Alias {alias} unset.",
                                   5000)
                    except KeyError:
                        pass
                else:
                    try:
                        Nav.conf["aliases"][alias] = actual
                    except KeyError:
                        Nav.conf["aliases"] = {alias: actual}
                    logger.debug(f"Alias: {alias} = {actual}")
                    Pub.notify("App", f"{self.pid}: Alias {alias} set to "
                               f"{actual}.", 5000)
                self.abar.setText(self.location)
                return
            else:  # resolve alias
                try:
                    loc = Nav.conf["aliases"][loc]
                except KeyError:
                    logger.error(f"Alias {loc} not set")
                    Pub.notify("App", f"{self.pid}: Alias {loc} not set.")
                    self.abar.setText(self.location)
                    return
        ret = self.tabbar.currentWidget().navigate(loc)
        self.sb.showMessage(self.tabbar.currentWidget().status_info)
        if ret:
            self.location = loc
        else:
            self.abar.setText(self.location)

    def tab_changed(self):
        """Update GUI elements to reflect the tab change."""
        self.update_gui(self.tabbar.currentWidget().location)

    def position_tree(self, loc: str):
        """Repositions tree based on the currently navigated folder."""
        index = self.tree.model.index(loc, 0)
        self.tree.setCurrentIndex(index)
        self.tree.scrollTo(index)

    def update_gui(self, loc):
        """Updates GUI to sync with navigations."""
        if loc != self.location:
            try:
                NavWatcher.remove_path(self.location, self.change_detected)
                NavWatcher.add_path(loc, self.change_detected)
                os.chdir(loc)
                self.location = loc
                self.abar.setText(self.location)
                self.position_tree(self.location)
            except (FileNotFoundError, NotADirectoryError):
                logger.error(f"{loc} not found")
        else:
            self.abar.setText(loc)
        # load tab if not already loaded
        self.tabbar.currentWidget().load_tab()
        self.tabbar.currentWidget().bcbar.create_crumbs(loc)
        self.check_navigation_options()
        self.filter_edit.setText(self.tabbar.currentWidget().filter_text)
        self.sb.showMessage(self.tabbar.currentWidget().status_info)
        self.pane_updated.emit(self)

    def check_navigation_options(self):
        """Checks if can navigate up/back/foreward"""
        self.can_go_up = True if self.tabbar.currentWidget().location \
            != os.sep else False
        self.can_go_back = True if self.tabbar.currentWidget().history \
            else False
        self.can_go_forward = True if self.tabbar.currentWidget().future \
            else False

    def update_status_bar(self, msg):
        """Updates status bar for pane with provided message."""
        self.sb.showMessage(msg)



import os
import pathlib
import sys
import time
from PyQt5 import QtCore, QtWidgets, QtGui
from PIL import Image
from PIL.ImageQt import ImageQt
from .core import NavStates, NavView
from .helper import logger, humansize, to_bytes
from .pub import Pub


class NavIcon:
    """Icon store for files."""
    icon_map = {}
    iconProvider = QtWidgets.QFileIconProvider()

    @classmethod
    def get_icon(cls, path, ext=None):
        """Return icon from cache or pull it into the cache and return."""
        if ext is None:
            if os.path.isfile(path):
                ext = pathlib.Path(path).suffix
        if ext not in cls.icon_map:
            file_info = QtCore.QFileInfo(path)
            icon = cls.iconProvider.icon(file_info)
            if ext is None:
                return icon
            cls.icon_map[ext] = icon
        return cls.icon_map[ext]


class NavItemModel(QtCore.QAbstractItemModel):
    """Custom File System Model for this application."""
    tw = th = 64

    def __init__(self, parent, header, *args, mylist=[]):
        super().__init__(parent, *args)
        self.parent = parent
        self.pid = self.parent.pid
        self.header = header
        self.files = []
        self.fcount = self.dcount = self.total = 0
        self.last_read = 0
        self.state = -1

    def model_size(self, width, height):
        """Set the size for icons and thumbnails."""
        self.tw = width
        self.th = height

    def list_dir(self, d: str):
        """Updates the model with director listing."""
        logger.debug(f"Invoked by {sys._getframe().f_back.f_code.co_name}")
        if not os.path.exists(d):
            self.files = []
            Pub.notify(f"App.{self.pid}.Tabs",
                       f"{self.pid}: {d} does not exist")
            return
        try:
            os.chdir(d)
        except PermissionError:
            self.status_info = f"Permission Denied for {d}"
            return False
        except NotADirectoryError:
            logger.error(f"{d} is not a directory")
            return False

        self.files = []
        self.fcount = self.dcount = self.total = self.selsize = 0
        cur_sel = []
        logger.debug(f"Listing {d} - {len(cur_sel)} selected")
        self.layoutAboutToBeChanged.emit()
        with os.scandir(d) as it:
            self.last_read = os.stat(d).st_mtime
            for entry in it:
                if entry.is_file():
                    state = 0
                    ext = pathlib.Path(entry.name).suffix.lstrip('.')
                    self.fcount += 1
                    size = humansize(entry.stat().st_size)
                else:
                    state = NavStates.IS_DIR
                    ext = None
                    self.dcount += 1
                    size = None
                modified = str(time.strftime('%Y-%m-%d %H:%M',
                               time.localtime(entry.stat().st_mtime)))
                self.total += entry.stat().st_size
                if entry.name in cur_sel:
                    state |= NavStates.IS_SELECTED
                self.files.append([entry.name,
                                   ext, size, modified, state, ])
        self.layoutChanged.emit()

    def insert_row(self, new_item: str):
        """Inserts a new item to the model."""
        if new_item not in self.files:
            try:
                if os.path.isfile(new_item):
                    state = 0
                    ext = pathlib.Path(new_item).suffix.lstrip('.')
                    self.fcount += 1
                    stats = os.lstat(new_item)
                    size = humansize(stats.st_size)
                    self.total += stats.st_size
                else:
                    state = NavStates.IS_DIR
                    ext = None
                    self.dcount += 1
                    size = ''
                    stats = os.lstat(new_item)
                modified = str(time.strftime('%Y-%m-%d %H:%M',
                               time.localtime(stats.st_mtime)))
                new_pos = self.rowCount()
                self.beginInsertRows(QtCore.QModelIndex(), new_pos, new_pos)
                self.files.append([new_item, ext, size, modified, state])
                Pub.notify("App", f"{self.pid}: {new_item} was added.")
                self.endInsertRows()
                return True
            except FileNotFoundError:
                pass

    def update_row(self, upd_item: str):
        """Updates a row in the model."""
        for item in self.files:
            if item[0] == upd_item:
                ind = self.files.index(item)
                if item[self.state] & ~NavStates.IS_DIR:
                    try:
                        size = to_bytes(item[2])
                        self.total -= size
                    except Exception:
                        logger.debug(f"Error getting size for {upd_item}",
                                     exc_info=True)
                stats = os.lstat(upd_item)
                size = humansize(stats.st_size)
                self.total += stats.st_size
                self.layoutAboutToBeChanged.emit()
                self.files[ind][2] = size
                self.files[ind][3] = str(time.strftime('%Y-%m-%d %H:%M',
                                         time.localtime(stats.st_mtime)))
                self.layoutChanged.emit()
                self.last_read = os.stat(self.parent.location).st_mtime

                Pub.notify("App", f"{self.pid}: {upd_item} was modified.")
                return

    def rename_row(self, old_name: str, new_name: str):
        """Renames a row in the model."""
        for item in self.files:
            if item[0] == old_name:
                ind = self.files.index(item)
                self.files[ind][0] = new_name
                self.dataChanged.emit(self.createIndex(0, 0),
                                      self.createIndex(self.rowCount(0),
                                      self.columnCount(0)))
                self.last_read = os.stat(self.parent.location).st_mtime
                Pub.notify("App", f"{self.pid}: {old_name} was renamed to "
                           f"{new_name}.")
                return

    def remove_row(self, rem_item: str):
        """ Remove a row from the model."""
        for item in self.files:
            if rem_item == item[0]:
                if item[self.state] & NavStates.IS_DIR:
                    self.dcount -= 1
                else:
                    try:
                        size = to_bytes(item[2])
                        self.total -= size
                    except Exception:
                        logger.debug(f"Error getting size for {rem_item}",
                                     exc_info=True)
                    self.fcount -= 1
                if item[self.state] & NavStates.IS_SELECTED:
                    self.selcount -= 1
                    try:
                        self.selsize -= size
                    except UnboundLocalError:
                        pass
                index = self.files.index(item)
                self.beginRemoveRows(QtCore.QModelIndex(), index, index)
                self.files.pop(index)
                self.endRemoveRows()
                Pub.notify("App", f"{self.pid}: {rem_item} was deleted.")
                break
        return True

    def rowCount(self, parent=None):
        """Returns the no. of rows in current model."""
        return len(self.files)

    def columnCount(self, parent=None):
        """Returns the no. of rows in current model."""
        return len(self.header)

    def index(self, row, column, parent=None):
        """Re-implemented to return index of a row."""
        return QtCore.QAbstractItemModel.createIndex(self, row, column, row)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        """Returns data to be displayed in the model."""
        if not index.isValid():
            return None
        try:
            row = index.row()
            column = index.column()
            value = self.files[row][column]
            if role == QtCore.Qt.EditRole:
                return value
            elif role == QtCore.Qt.DecorationRole:
                if column == 4 or (self.parent.vtype == NavView.Thumbnails):
                    try:
                        im = Image.open(f"{self.parent.location}{os.sep}"
                                        f"{self.files[row][0]}")
                        im.thumbnail((self.tw, self.th), Image.ANTIALIAS)
                        return QtGui.QImage(ImageQt(im))
                    except Exception:
                        # Icon if thumbnails can't be generated
                        return NavIcon.get_icon(self.files[row][0])
                elif column == 0:
                    return NavIcon.get_icon(self.files[row][0])
                else:
                    return None
            elif role == QtCore.Qt.DisplayRole:
                return value if column != 4 else ""
            elif role == QtCore.Qt.CheckStateRole and column == 0:
                if self.files[row][self.state] & NavStates.IS_SELECTED:
                    return QtCore.Qt.Checked
                else:
                    return QtCore.Qt.Unchecked
        except IndexError:
            pass

    def headerData(self, column, orientation, role):
        """Return caption for the headers."""
        if orientation == QtCore.Qt.Horizontal \
                and role == QtCore.Qt.DisplayRole:
            try:
                return self.header[column].caption
            except IndexError:
                return ""
        return None

    def flags(self, index):
        """Re-implemented to allow checkbox on the name column and
        prevent selection on other columns."""
        if not index.isValid():
            return QtCore.Qt.ItemIsEnabled
        if index.column() == 0:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable | \
                QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsEditable | \
                QtCore.Qt.ItemIsDragEnabled | QtCore.Qt.ItemIsDropEnabled
        else:
            return QtCore.Qt.ItemIsEnabled

    def setData(self, index, value, role):
        """Handles item (de)selection."""
        if not index.isValid():
            return False
        if role == QtCore.Qt.CheckStateRole:
            if value == QtCore.Qt.Checked:
                self.files[index.row()][self.state] |= NavStates.IS_SELECTED
                self.selsize += to_bytes(self.files[index.row()][2])
            else:
                self.files[index.row()][self.state] &= ~NavStates.IS_SELECTED
                self.selsize -= to_bytes(self.files[index.row()][2])
            # Emit signal to select row only if not invoked by it
            if sys._getframe().f_back.f_code.co_name == "__init__":
                self.dataChanged.emit(index, index)
            return True
        elif role == QtCore.Qt.EditRole:
            r = index.row()
            c = index.column()
            if c == 0:
                old = self.files[r][c]
                logger.debug(f"Rename {old} to {value}")
                self.rename(old, value)
        # self.dataChanged.emit(index, index)
        return False

    def rename(self, old, new):
        if os.path.exists(f"{new}"):
            logger.error(f"{self.parent.location}: {new} - Exists")
            Pub.notify("App", f"{self.pid}: {new} - exists.")
        else:
            logger.info(f"{self.parent.location}: Rename {old} to {new}")
            try:
                os.rename(old, new)
                Pub.notify(f"App.{self.pid}.Tab.Files.Renamed",
                           f"{self.pid}: {old} renamed to {new}")
            except OSError:
                logger.error(f"{self.parent.location}: Error renaming "
                             f"from {old} to {new}", exc_info=True)
                Pub.notify("App", f"{self.pid}: Error renaming.")

    def get_selection_stats(self):
        """Get stats of selected items"""
        self.selsize = self.selcount = 0
        for item in self.files:
            if item[self.state] & NavStates.IS_SELECTED & ~NavStates.IS_DIR:
                self.selsize += to_bytes(item[2])
                self.selcount += 1
        return self.selcount, self.selsize



import collections
import os
import pathlib
import shutil
import stat
import subprocess
import sys
import threading
from PyQt5 import QtWidgets, QtCore, QtGui
from send2trash import send2trash
from .breadcrumbs import NavBreadCrumbsBar
from .helper import logger, humansize
from .pub import Pub
from .model import NavItemModel
from .custom import (NavSortFilterProxyModel, NavHeaderView, NavColumn)
from .core import Nav, NavView, NavSize


class NavTabWidget(QtWidgets.QTabWidget):
    """Re-implemented to present tab menu"""
    cMenu = None  # common contextMenu across class
    tab_created = QtCore.pyqtSignal()

    def __init__(self, parent, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parent = parent
        self.pid = self.parent.pid
        if self.cMenu is None:
            # logger.debug("Creating context menu for tabbar")
            self.__class__.cMenu = QtWidgets.QMenu()
            items = [
                Nav.actions["new_tab"], Nav.actions["close_tab"],
                Nav.actions["rename_tab"],
                Nav.actions["close_other_tabs"],
                Nav.actions["close_left_tabs"],
                Nav.actions["close_right_tabs"],
            ]
            Nav.build_menu(self, items, self.cMenu)

    def new_tab(self, tab_info={}):
        """Creates a new tab with default/provided values."""
        t = NavTab(tab_info, parent_id=self.pid)
        t.location_changed.connect(self.update_gui_with_tab)
        i = self.addTab(t, tab_info["location"])
        self.set_caption(i, caption=t.caption, loc=t.location)
        self.tab_created.emit()  # install filter for new tabs

    def update_gui_with_tab(self, loc):
        """Updates GUI and sets tab caption to sync with navigations."""
        cur_tab = self.currentIndex()
        self.set_caption(cur_tab, loc=loc)
        # self.currentWidget().bcbar.create_crumbs(loc)
        self.currentChanged.emit(cur_tab)

    def set_caption(self, index: int=None, caption: str=None,
                    loc: str=None, rename: bool=False):
        """Renames a tab with provided/prompted caption."""
        if index is None:
            index = self.get_index()
        # logger.debug(f"Tab data: {self.tabBar().tabData(index)}")
        if not caption:
            if rename:
                current_caption = self.tabText(index)
                caption, ok = QtWidgets.QInputDialog.getText(
                                self, 'Rename Tab', 'Enter new caption:',
                                QtWidgets.QLineEdit.Normal, current_caption)
                if not ok:
                    return
                self.tabBar().setTabData(index, caption)
            if not caption:
                if loc is None:
                    loc = self.tabToolTip(index)
                if self.tabBar().tabData(index):
                    self.setTabToolTip(index, loc)
                    return
                caption = os.path.basename(loc)
                if caption == "":
                    caption = "/"
        else:
            self.tabBar().setTabData(index, caption)
        self.setTabText(index, caption)
        if loc is not None:
            self.setTabToolTip(index, loc)

    def contextMenuEvent(self, event):
        """Presents a context menu at mouse position."""
        tab = self.tabBar().tabAt(event.pos())
        if tab < 0:
            return
        self.cMenu.exec_(event.globalPos())
        logger.debug(f"Mouse is on tab# {self.tabBar().tabAt(event.pos())}")

    def get_index(self, depth=3):
        """Gets index for actioning tab."""
        if sys._getframe(depth).f_back.f_code.co_name == "__init__":
            return self.rc_on
        else:
            return self.currentIndex()

    def select_tab(self, index):
        """Select next/previous or the mentioned tab."""
        if index == "next":
            index = self.currentIndex() + 1
        elif index == "prev":
            index = self.currentIndex() - 1
        index = index % self.count()
        self.setCurrentIndex(index)

    def close_tab(self, index=None):
        """Close current/provided tab."""
        if self.count() == 1:
            self.setTabsClosable(False)
            Pub.notify(f"App.{self.pid}.Tabs",
                       f"{self.pid}: Can't close last tab.")
            return
        else:
            self.setTabsClosable(True)
        if index is None:
            index = self.get_index()
        widget = self.widget(index)
        if widget is not None:
            widget.deleteLater()
        self.removeTab(index)
        logger.debug(f"Removed tab {index}")

    def close_other_tabs(self, index=None):
        """Closes all tabs except this."""
        if index is None:
            index = self.get_index()
        for ind in range(self.count(), -1, -1):
            if ind != index:
                self.close_tab(ind)
        Pub.notify(f"App.{self.pid}.Tabs",
                   f"{self.pid}: All other tabs closed")

    def close_left_tabs(self, index=None):
        """Close tabs to the left."""
        if index is None:
            index = self.get_index()
        for ind in range(index-1, -1, -1):
            self.close_tab(ind)
        Pub.notify(f"App.{self.pid}.Tabs", f"{self.pid}: Left tabs closed")

    def close_right_tabs(self, index=None):
        """Close tabs to the right."""
        if index is None:
            index = self.get_index()
        for ind in range(self.count(), index, -1):
            self.close_tab(ind)
        Pub.notify(f"App.{self.pid}.Tabs", f"{self.pid}: Right tabs closed")

    def mousePressEvent(self, event):
        """Re-implement to capture the index at mouse position."""
        if event.button() == QtCore.Qt.RightButton:
            index = self.tabBar().tabAt(event.pos())
            self.rc_on = index
        else:
            super().mousePressEvent(event)


class NavTab(QtWidgets.QFrame):
    """Class to handle base skeleton for tabs."""
    location_changed = QtCore.pyqtSignal(str)
    status_updated = QtCore.pyqtSignal(str)
    widths = heights = {
        NavSize.Tiny: 16, NavSize.Small: 64, NavSize.Medium: 96,
        NavSize.Large: 128, NavSize.XL: 256
    }

    def __init__(self, tab_info, parent_id):
        self.mutex = QtCore.QMutex()
        super().__init__()
        self.pid = parent_id
        self.bcbar = NavBreadCrumbsBar("/")
        self.bcbar.clicked.connect(self.navigate)
        self._offset = QtCore.QPoint(30, 30)
        self.history = collections.deque(maxlen=64)
        self.future = collections.deque(maxlen=64)
        try:
            self.history = tab_info["history"]
        except (KeyError, TypeError):
            pass
        try:
            self.future = tab_info["future"]
        except (KeyError, TypeError):
            pass
        try:
            self.location = tab_info["location"]
        except (KeyError, TypeError):
            self.location = str(pathlib.Path.home())
            tab_info["location"] = self.location
        try:
            self.caption = tab_info["caption"]
        except (KeyError, TypeError):
            self.caption = None
        self.status_info = ''
        self.filter_text = ''
        self._loading = False
        try:
            self.sort_column = int(tab_info["sort_column"])
        except (KeyError, TypeError):
            self.sort_column = 0
        try:
            self.sort_order = int(tab_info["sort_order"])
        except (KeyError, TypeError):
            self.sort_order = 0

        self.header = [
            NavColumn("Name", 300),
            NavColumn("Ext", 50),
            NavColumn("Size", 100),
            NavColumn("Modified", 150),
            NavColumn("Thumbnails", 128)
        ]
        self.model = NavItemModel(self, self.header)
        self.proxy = NavSortFilterProxyModel(self)
        self.proxy.setSourceModel(self.model)
        self.proxy.setFilterKeyColumn(0)
        self.lv = self.tv = None
        self.init_table_view()
        self.init_list_view()
        self.view = self.tv
        self.rubberBand = QtWidgets.QRubberBand(
            QtWidgets.QRubberBand.Rectangle, self.view.viewport())

        # Show columns for tab
        if "columns" in tab_info:
            for idx, h in enumerate(self.header):
                flag = 0
                for col in tab_info["columns"]:
                    if h.caption == col[0]:
                        flag = 1
                        break
                if not flag and idx != 0:
                    self.hv.hideSection(idx)
                    h.visible = False
                elif col[2] != -1:
                    vis_ind = self.hv.visualIndex(idx)
                    self.hv.moveSection(vis_ind, col[2])
            # Capture movements made before connecting the signal
            self.columns_visibility_changed(4, "Thumbnails",
                                            self.header[4].visible)
            self.columns_moved(0, 0, 0)
        # Connect to save column movements
        self.hv.sectionMoved.connect(self.columns_moved)
        self.hv.visibility_changed.connect(self.columns_visibility_changed)
        self.lyt = QtWidgets.QVBoxLayout()
        self.lyt.setSpacing(0)
        self.lyt.setContentsMargins(0, 0, 0, 0)
        self.lyt.addWidget(self.bcbar)
        self.lyt.addWidget(self.view)
        self.setLayout(self.lyt)
        self.install_filters()
        self.proxy.dataChanged.connect(self.row_sel)
        try:
            self.vtype = tab_info["view"]
        except KeyError:
            self.vtype = NavView.Details
        try:
            self.vsize = tab_info["itsize"]
        except KeyError:
            self.vsize = NavSize.Tiny
        self.switch_view(self.vtype, self.vsize)

    def init_header(self):
        """Initializes a header with checkbox selection."""
        self.hv = NavHeaderView(self.header)
        self.hv.setSectionsMovable(True)
        self.hv.setSectionsClickable(True)
        self.hv.setHighlightSections(True)
        self.hv.clicked.connect(self.updateModel)
        self.hv.setModel(self.model)

    def init_table_view(self):
        """Initialize Details View."""
        self.tv = QtWidgets.QTableView()
        self.init_header()
        self.tv.setWordWrap(False)
        self.tv.setSelectionBehavior(self.tv.SelectItems)
        self.tv.setEditTriggers(QtWidgets.QAbstractItemView.EditKeyPressed)
        self.tv.setHorizontalHeader(self.hv)
        self.tv.setSortingEnabled(True)
        self.tv.setAlternatingRowColors(True)
        self.tv.setModel(self.proxy)
        self.tv.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.tv.horizontalHeader().sortIndicatorChanged.connect(
                                                    self.sortIndicatorChanged)
        self.tv.SelectionBehavior(1)
        self.tv.SelectionMode(7)
        self.tv.selectionModel().selectionChanged.connect(self.rows_selected)
        self.tv.doubleClicked.connect(self.double_clicked)
        self.tv.setDragDropMode(
                QtWidgets.QAbstractItemView.DragDrop &
                ~QtWidgets.QAbstractItemView.InternalMove)

    def init_list_view(self):
        """Initialize List View."""
        self.lv = QtWidgets.QListView()
        self.lv.setWordWrap(True)
        self.lv.setTextElideMode(QtCore.Qt.ElideMiddle)
        self.lv.setMovement(QtWidgets.QListView.Snap)
        self.lv.setResizeMode(QtWidgets.QListView.Adjust)
        self.lv.setDragEnabled(True)
        self.lv.setDragDropMode(QtWidgets.QAbstractItemView.DragDrop &
                                ~QtWidgets.QAbstractItemView.InternalMove)
        self.lv.setSelectionBehavior(self.lv.SelectItems)
        self.lv.setEditTriggers(QtWidgets.QAbstractItemView.EditKeyPressed)
        self.lv.setAlternatingRowColors(True)
        self.lv.setModel(self.proxy)
        self.lv.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.lv.setUniformItemSizes(True)
        self.lv.SelectionBehavior(1)
        self.lv.SelectionMode(7)
        self.lv.selectionModel().selectionChanged.connect(self.rows_selected)
        self.lv.doubleClicked.connect(self.double_clicked)

    def switch_view(self, new_view, size=NavSize.Tiny):
        """Switch between views"""
        self.vtype = new_view
        self.vsize = size
        width = self.widths[size]
        height = self.heights[size]
        self.model.model_size(width, height)

        if new_view == NavView.Details:
            if self.tv is not self.view:
                selections = self.view.selectionModel().selection()
                self.lyt.removeWidget(self.view)
                self.view.setGeometry(0, 0, 0, 0)
                self.view = self.tv
                self.lyt.addWidget(self.view)
                self.view.selectionModel().select(
                    selections, QtCore.QItemSelectionModel.ClearAndSelect)
            return
        elif self.lv is not self.view:
            selections = self.view.selectionModel().selection()
            self.lyt.removeWidget(self.view)
            self.view.setGeometry(0, 0, 0, 0)
            self.view = self.lv
            self.lyt.addWidget(self.view)
            self.view.selectionModel().select(
                selections, QtCore.QItemSelectionModel.ClearAndSelect)
        if new_view == NavView.List:
            self.view.setViewMode(QtWidgets.QListView.ListMode)
            self.lv.verticalScrollBar().setSingleStep(height)
            self.lv.setIconSize(QtCore.QSize(width, height))
            # self.lv.setGridSize(QtCore.QSize(300, 30))
            # self.lv.setSpacing(5)
            self.lv.setWrapping(True)
        elif new_view in [NavView.Icons, NavView.Thumbnails]:
            self.view.setViewMode(QtWidgets.QListView.IconMode)
            self.lv.setIconSize(QtCore.QSize(width, height))
            self.lv.setGridSize(QtCore.QSize(width+10, height+30))
            self.lv.verticalScrollBar().setSingleStep(height+30)

    def install_filters(self):
        """Install event filter in all children of the panel."""
        for widget in self.findChildren(QtWidgets.QWidget):
            widget.installEventFilter(self)

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, value):
        if value == "":
            return
        self._location = value

    def set_filter(self, filter_text):
        """Apply the filter provided in filter box."""
        self.filter_text = filter_text
        self.proxy.setFilterCaseSensitivity(False)
        self.proxy.setFilterRegExp(filter_text)
        for i in range(self.proxy.rowCount()):
            index = self.proxy.index(i, 0)
            if index not in self.view.selectionModel().selectedIndexes():
                self.proxy.setData(index, QtCore.Qt.Unchecked,
                                   QtCore.Qt.CheckStateRole)

    def eventFilter(self, obj, event):
        """Reimplemented to handle active pane."""
        if self.vtype == NavView.Details:
            if event.type() == QtCore.QEvent.MouseButtonRelease:
                self.tv_mouseReleaseEvent(event)
            elif event.type() == QtCore.QEvent.MouseButtonPress:
                self.tv_mousePressEvent(event)
            elif event.type() == QtCore.QEvent.MouseMove:
                self.tv_mouseMoveEvent(event)
        return False

    def tv_mousePressEvent(self, event):
        """Reimplemented for custom handling."""
        self.origin = event.pos()
        self._modifier = event.modifiers()
        if self.view.indexAt(self.origin).column() != 0:
            self.view.setDragDropMode(QtWidgets.QAbstractItemView.DropOnly)
            self._selection = self.view.selectionModel().selection()
            self.rubberBand.setGeometry(QtCore.QRect(self.origin,
                                                     QtCore.QSize()))
            self.rubberBand.show()
        else:
            self.view.setDragDropMode(
                QtWidgets.QAbstractItemView.DragDrop &
                ~QtWidgets.QAbstractItemView.InternalMove)
        # super().mousePressEvent(event)

    def tv_mouseMoveEvent(self, event):
        """Reimplemented for custom handling."""
        if self.rubberBand.isVisible():
            pos = event.pos()
            self.rubberBand.setGeometry(
                QtCore.QRect(self.origin, pos).normalized())
            rect = self.rubberBand.geometry()
            tl = rect.topLeft()
            br = rect.bottomRight()
            qis = QtCore.QItemSelection(self.view.indexAt(tl),
                                        self.view.indexAt(br))
            mode = QtCore.QItemSelectionModel.ClearAndSelect
            if self._modifier == QtCore.Qt.ControlModifier:
                mode = QtCore.QItemSelectionModel.Clear | \
                       QtCore.QItemSelectionModel.Toggle
                qis.merge(self._selection, mode)
            if self._modifier == QtCore.Qt.ShiftModifier:
                qis.merge(self._selection, mode)
            self.view.selectionModel().select(qis, mode)
            event.accept()

    def tv_mouseReleaseEvent(self, event):
        """Reimplement for custom handling."""
        # Unset selection if selection rectangle was drawn and hide rectangle
        if self.rubberBand.isVisible():
            self.rubberBand.hide()
            self._selection = None
            event.accept()
        else:
            super().mouseReleaseEvent(event)

    def double_clicked(self, index):
        """Get details to open the double clicked item."""
        if index == self.view.currentIndex():
            item = self.proxy.itemData(index).get(0)
        else:
            item = None
        if item:
            loc = os.path.join(self.location, item)
        else:
            loc = os.path.dirname(self.location)
        self.opener(loc)

    def opener(self, loc):
        """Launches the file or navigates to the said location."""
        if os.path.isdir(loc):
            self.navigate(loc)
        else:
            if os.name == 'posix':  # For Linux, Mac, etc.
                subprocess.call(('xdg-open', loc))
            elif sys.platform.startswith('darwin'):
                subprocess.call(('open', loc))
            elif os.name == 'nt':  # For Windows
                os.startfile(loc)

    def navigate(self, d):
        """Navigates to the provided location."""
        logger.debug(f"Invoked by {sys._getframe().f_back.f_code.co_name}")
        logger.debug(f"Navigating to {d}")
        if d != self.location:
            if Nav.conf["history_without_dupes"] and \
                    self.location in self.history:
                self.history.remove(self.location)
            self.history.append(self.location)
            self.load_tab(d)
            self.future.clear()

    def go_up(self):
        """Go up the tree."""
        try:
            loc = os.path.dirname(self.location)
            self.opener(loc)
        except OSError:
            logger.error(exc_info=True)

    def go_back(self):
        """Go back in history."""
        try:
            d = self.history.pop()
            if d != self.location:
                if self.location in self.future:
                    self.future.remove(self.location)
                self.future.append(self.location)
                self.load_tab(d)
                logger.debug(f"Future: {self.future}")
        except IndexError:
            logger.error(f"No more back")

    def go_forward(self):
        """Go forward in future."""
        try:
            d = self.future.pop()
            if d != self.location:
                if self.location in self.history:
                    self.history.remove(self.location)
                self.history.append(self.location)
                self.load_tab(d)
                logger.debug(f"History: {self.history}")
        except IndexError:
            logger.error(f"No more forward")

    def latest_history(self):
        return self.history[0]

    def row_sel(self, a, b):
        """Toggle row selection on checkbox click."""
        self.view.selectionModel().select(a, QtCore.QItemSelectionModel.Toggle)

    def model_changed(self, a, b):
        logger.debug(f"called {a} {b}")

    def rows_selected(self, sel, desel):
        """Handle row (de)selections."""
        for index in sel.indexes():
            self.proxy.setData(index, QtCore.Qt.Checked,
                               QtCore.Qt.CheckStateRole)
        for index in desel.indexes():
            if not self.proxy.itemData(index):
                continue

            if self.filter_text == "":
                self.proxy.setData(index, QtCore.Qt.Unchecked,
                                   QtCore.Qt.CheckStateRole)

        selstat = len(self.view.selectionModel().selectedIndexes())
        if selstat:
            selcount, selsize = self.model.get_selection_stats()
            selinfo = f" Selected: {selcount} : " \
                      f"{humansize(selsize)}"
            if (selstat >= self.proxy.rowCount()):
                self.hv.updateCheckState(1)
            else:
                self.hv.updateCheckState(2)
        else:
            selinfo = ""
            self.hv.updateCheckState(0)
        Pub.notify(f"Panes.{self.pid}.Tabs", f"{self.status_info} {selinfo}")

    def updateModel(self, index, state):
        if index != 0:
            return
        if state:
            if self.vtype == NavView.Details:
                self.view.selectColumn(index)
            else:
                self.view.selectAll()
        else:
                self.view.clearSelection()

    def invert_selection(self):
        """Toggles (de)selection of items in current listing."""
        for i in range(self.proxy.rowCount()):
            ix = self.proxy.index(i, 0)
            self.view.selectionModel().select(
                ix, QtCore.QItemSelectionModel.Toggle)

    def keyPressEvent(self, event):
        """Reimplemented keyPressEvent for custom handling."""
        if event.key() == QtCore.Qt.Key_Return:
            index = self.view.currentIndex()
            item = self.proxy.itemData(index).get(0)
            if self.view.state() != QtWidgets.QAbstractItemView.EditingState:
                loc = os.path.join(self.location, item)
                self.opener(loc)
        elif event.key() == QtCore.Qt.Key_Backspace:
            loc = os.path.dirname(self.location)
            self.navigate(loc)
        else:
            super().keyPressEvent(event)
            event.ignore()

    def load_tab(self, loc=None, forced=False):
        """Loads a tab if not already loaded or is stale."""
        logger.debug(f"Invoked by {sys._getframe().f_back.f_code.co_name}")
        if loc is None:
            loc = self.location
        if not os.path.isdir(loc):
            logger.debug(f"{loc} isn't a directory")
            Pub.notify("App", f"{loc} isn't a directory.")
            return
        if (forced and self._loading is False) or loc != self.location \
                or self.model.last_read < os.stat(loc).st_mtime:
            if not forced:
                self.filter_text = ""
            self.hv.updateCheckState(0)  # Uncheck main checkbox
            self._loading = True
            self.model.list_dir(loc)
            try:
                free_disk = humansize(shutil.disk_usage(loc)[2])
            except OSError:
                free_disk = ""
            self.status_info = f"Files: {self.model.fcount}, Dirs: " \
                f"{self.model.dcount} Total: {humansize(self.model.total)} " \
                f"Free: {free_disk}"
            if self.vtype == NavView.Details:
                self.tv.sortByColumn(self.sort_column, self.sort_order)
            self._loading = False
            if self.location != os.path.abspath(loc):
                self.view.clearSelection()
                self.location = os.path.abspath(loc)
                self.location_changed.emit(self.location)

    def change_detected(self, evt):
        """Invokes appropriate methods to add/update/remove listed files."""
        logger.debug(f"Invoked by {sys._getframe().f_back.f_code.co_name}")
        name = os.path.basename(evt.src_path)
        logger.debug(f"{name} {evt.event_type}")
        if evt.event_type == "deleted":
            logger.debug(f"{evt.src_path} deleted in {self.location}")
            self.model.remove_row(name)
        elif evt.event_type == "created":
            logger.debug(f"{evt.src_path} created in {self.location}")
            self.model.insert_row(name)
        elif evt.event_type == "moved":
            logger.debug(f"{self.location}: Moved {evt.src_path} to "
                         f"{evt.dest_path}")
            new_name = os.path.basename(evt.dest_path)
            self.model.rename_row(name, new_name)
        elif evt.event_type == "modified":
            logger.debug(f"{self.location}: {evt.src_path} Modified")
            self.model.update_row(name)
        # selstat = len(self.view.selectionModel().selectedIndexes())
        selstat = len(self.view.selectionModel().selectedIndexes())
        if selstat:
            selcount, selsize = self.model.get_selection_stats()
            selinfo = f" Selected: {selcount} : " \
                      f"{humansize(selsize)}"
        else:
            selinfo = ""
        self.status_info = f"Files: {self.model.fcount}, Dirs: " \
            f"{self.model.dcount} Total: {humansize(self.model.total)} " \
            f"Free: {humansize(shutil.disk_usage(self.location)[2])} {selinfo}"

    def sortIndicatorChanged(self, logicalIndex, sortOrder):
        """Remembers the sorted column and order."""
        self.sort_column = logicalIndex
        self.sort_order = sortOrder

    def rename_file(self):
        """Invokes inline renaming of the current file."""
        index = self.view.currentIndex()
        if not index:
            return
        self.view.edit(index)

    def trash(self):
        """Delete the current list of selected files to trash."""
        files = [self.proxy.itemData(index).get(0)
                 for index in self.view.selectionModel().selectedIndexes()]
        for f in files:
            fullname = os.path.join(self.location, f)
            try:
                send2trash(fullname)
                self.model.remove_row(f)
            except OSError:
                logger.error(f"Error deleting file", exc_info=True)
                Pub.notify("App", f"{self.pid}: Error deleting file",
                           exc_info=True)

    def delete(self):
        """Permanently delete the currently selected files."""
        files = [os.path.join(self.location, self.proxy.itemData(index).get(0))
                 for index in self.view.selectionModel().selectedIndexes()]
        for f in files:
            try:
                if os.path.isdir(f):
                    shutil.rmtree(f, onerror=self.remove_readonly)
                else:
                    os.unlink(f)
            except FileNotFoundError:
                logger.error(f"{f} not found")
                Pub.notify(f"App", f"{self.pid}: {f} not found.")

    def del_dir_up(self):
        """Deletes the current directory and goes up one directory."""
        d = self.location
        new_location = str(pathlib.PurePath(d).parent)
        try:
            send2trash(d)
            self.navigate(new_location)
        except FileNotFoundError:
            logger.error(f"{d} not found")
            Pub.notify("App.{self.pid}.Tab", f"{self.pid}: {d} not found.")

    def remove_readonly(self, func, path, _):
        "Clear the readonly bit and reattempts the removal."
        os.chmod(path, stat.S_IWRITE)
        func(path)

    def new_file(self, kind):
        """Creates new file or folder."""
        kind = kind.title()
        if kind == "Folder":
            filename = "new_folder"
        else:
            filename = "new_file"
        inc = ''
        while os.path.exists(filename + str(inc)):
            if inc:
                inc = f"({int(inc[1:-1])+1})"
            else:
                inc = "(1)"
        filename = f"{filename}{inc}"
        try:
            if kind == "Folder":
                os.makedirs(filename)
            else:
                os.mknod(filename)
            Pub.notify("App", f"{self.pid}: {kind} - {filename} created")
        except OSError:
            logger.error(f"Error creating {filename}", exc_info=True)
            Pub.notify("App", f"{self.pid}: Error creating {filename}")

    def copy(self, cut=False):
        """Cut/Copy files to clipboard."""
        files = [QtCore.QUrl.fromLocalFile(
            os.path.join(self.location, self.proxy.itemData(index).get(0)))
                 for index in self.view.selectionModel().selectedIndexes()]
        mime_data = self.proxy.mimeData(self.view.selectionModel().
                                        selectedIndexes())
        if cut:
            data = b'1'  # same as QtCore.QByteArray(0, '1')
            mime_data.setData("application/x-kde-cutselection", data)
            data = b'cut'
            mime_data.setData("x-special/gnome-copied-files", data)
        mime_data.setUrls(files)
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setMimeData(mime_data)

    def paste(self):
        """Pastes files from clipboard."""
        clipboard = QtWidgets.QApplication.clipboard()
        # check if cut or copy
        # x-kde-cutselection: is 1 if cut else 0
        # x-special/gnome-copied-files: has cut or copy mentioned
        logger.debug(clipboard.mimeData().formats())
        gnome_op = clipboard.mimeData().data(
            'x-special/gnome-copied-files').split(b'\n')[0]
        gnome_cut = True if gnome_op == b'cut'else False
        kde_op = clipboard.mimeData().data('application/x-kde-cutselection')
        kde_cut = True if kde_op == b'1' else False
        cut = True if kde_cut or gnome_cut else False
        logger.debug(f"Files were cut: {cut}")
        urls = [QtCore.QUrl.toLocalFile(url)
                for url in clipboard.mimeData().urls()]
        logger.debug(f"Paste {urls}")
        if not urls:
            return

        if cut:
            act = "move"
        else:
            act = "copy"
        self.t = threading.Thread(target=self.copier,
                                  args=(act, urls, self.location))
        self.t.start()

    def copier(self, act, sources, dest):
        """Invokes the copier in separate process."""
        Nav.copy_jobs += 1
        process = subprocess.Popen(
                    ['python3', Nav.copier, act] + sources + [dest])
        process.wait()
        Nav.copy_jobs -= 1
        logger.info(f"{act} {sources} {dest} -> completed")

    def startDrag(self, supported_actions):
        """Reimplemented to handle drag."""
        drag = QtGui.QDrag(self)
        t = [QtCore.QUrl.fromLocalFile(
                f"{self.location}{os.sep}{self.proxy.itemData(index).get(0)}")
             for index in self.view.selectionModel().selectedIndexes()]
        mime_data = self.proxy.mimeData(self.view.selectionModel().
                                        selectedIndexes())
        mime_data.setUrls(t)
        logger.debug(f"Dragging: {mime_data.urls()}")
        drag.setMimeData(mime_data)
        drag.exec_()

    def dragEnterEvent(self, event):
        """Reimplemented to handle drag"""
        logger.debug("startDrag")
        if self.rubberBand.isVisible:
            logger.debug("return")
            event.accept()
            return
        m = event.mimeData()
        if m.hasUrls():
            logger.debug(f"{event.mimeData().urls()}")
            event.accept()
            return
        event.ignore()

    def dragMoveEvent(self, event):
        """Reimplemented to handle drag"""
        # event.setDragDropMode(QtGui.QAbstractItemView.InternalMove)
        event.setDropAction(QtCore.Qt.CopyAction)
        event.accept()

    def dropEvent(self, event):
        """Reimplemented to handle drop"""
        m = event.mimeData()
        if m.hasUrls():
            logger.debug(f"Dropping urls: {m.urls()}")
            links = []
            drop_loc = self.proxy.data(self.view.indexAt(event.pos()))
            if drop_loc is None:
                drop_loc = self.location
            else:
                drop_loc = os.path.join(self.location, drop_loc)
                # if dropped on file, go back to parent folder
                if not os.path.isdir(drop_loc):
                    drop_loc = self.location
            logger.debug(f"Drop Location: {drop_loc}")
            copylist = []
            for url in m.urls():
                links.append(str(url.toLocalFile()))

            for link in links:
                logger.debug(f"Copying {link} to {drop_loc}")
                try:
                    basename = os.path.basename(link)
                    os.rename(link, f"{drop_loc}{os.sep}{basename}")
                except OSError:
                    logger.error(f"{self.location}: It's on different mount. "
                                 "Copying", exc_info=True)
                    copylist.append(link)

            if copylist:
                self.t = threading.Thread(target=self.copier,
                                          args=("copy", copylist, drop_loc))
                self.t.start()

    def handledrop(self, links):
        """Reimplemented to handle drop"""
        for url in links:
            if os.path.exists(url):
                logger.debug(url)

    def get_selected_items(self):
        """Returns list of selected item."""
        files = [os.path.join(self.location, self.proxy.itemData(index).get(0))
                 for index in self.view.selectionModel().selectedIndexes()]
        return files

    def columns_moved(self, ind, old, new):
        """Capture the new visual indices for each of the columns."""
        for idx, h in enumerate(self.header):
            h.position = self.hv.visualIndex(idx)

    def columns_visibility_changed(self, idx, cap, visible):
        """Update row height if Thumnails column visibility is changed."""
        if cap == "Thumbnails":
            if visible:
                self.tv.verticalHeader().setDefaultSectionSize(128)
            else:
                self.tv.verticalHeader().setDefaultSectionSize(20)



import os
import pathlib
from enum import IntFlag
from PyQt5 import QtWidgets
# from .helper import logger


class Nav:
    """Data store to hold application state."""
    app = str(pathlib.Path(__file__).resolve())
    app_dir = os.path.dirname(app)
    app_data = f"{app_dir}{os.sep}..{os.sep}appdata"
    copier = os.path.join(app_dir, "navcopier.py")
    copy_jobs = 0
    conf_file = os.path.join(str(pathlib.Path.home()), "navgator.json")

    conf = {
        "panes": {"total": 4, "active": "Pane 1", },
        "window": {"main_tree": True, "statusbar": True},
        "history_without_dupes": True,
        "sort_folders_first": True,
        "watch_all_tabs": True,
        "shortcuts": {"back": "backspace", },
        "colors": {"bcbar": {"active": "blue", "inactive": "green"}},
    }

    @classmethod
    def getsizes(cls, k, min, defaults):
        """Helper function to provide sizes for splitter elements."""
        try:
            elem = len(defaults)
            for i in range(elem):
                if cls.conf["dims"][k][i] < min[i]:
                    cls.conf["dims"][k] = defaults
                    break
        except Exception:
            cls.conf["dims"][k] = defaults
        return cls.conf["dims"][k]

    @classmethod
    def build_menu(cls, parent, d: dict, menu: QtWidgets.QMainWindow.menuBar):
        """Build up the menu recursively."""
        if isinstance(d, list):
            for v in d:
                if isinstance(v, QtWidgets.QAction):
                    menu.addAction(v)
                elif isinstance(v, dict):
                    if "sm" in v:
                        m2 = menu.addMenu(v["caption"])
                        cls.build_menu(parent, v["sm"], m2)
        elif isinstance(d, dict):
            for k, v in d.items():
                if "sm" in v:
                    m2 = menu.addMenu(v["caption"])
                    cls.build_menu(parent, v["sm"], m2)


class NavStates(IntFlag):
    """Enum to store state of items."""
    IS_DIR = 1
    IS_SELECTED = 2


class NavView(IntFlag):
    """Enum for different Views"""
    Details = 1
    List = 2
    Icons = 3
    Thumbnails = 4
    # SmallIcons = 3
    # MediumIcons = 4
    # LargeIcons = 5
    # XLIcons = 6
    # SmallThumbs = 7
    # MediumThumbs = 8
    # LargeThumbs = 9
    # XLThumbs = 10


class NavSize(IntFlag):
    Tiny = 1
    Small = 2
    Medium = 3
    Large = 4
    XL = 5


import collections
import functools
import logging.config
import sys


lconf = {
        'version': 1,
        'disable_existing_loggers': True,
        'formatters': {
            'detailed': {
                'class': 'logging.Formatter',
                'format': '%(asctime)s - %(relativeCreated)d - %(levelname)s'
                          ' - %(funcName)s - %(lineno)d - %(message)s',
                'datefmt': '%Y-%m-%d %H:%M:%S'
            }
        },
        'handlers': {
            'console': {
                'class': 'logging.StreamHandler',
                'level': 'DEBUG',
                'formatter': 'detailed',
            },
            'file': {
                'class': 'logging.FileHandler',
                'filename': '/tmp/navgator.log',
                'mode': 'w',
                'level': 'DEBUG',
                'formatter': 'detailed',
            },
        },
        'loggers': {
            'foo': {
                'handlers': []
            }
        },
        'root': {
            'level': 'DEBUG',
            'handlers': ['console', 'file']
        },
    }
logging.config.dictConfig(lconf)
logger = logging.getLogger(__name__)


def humansize(num, power=1024, sep=' ', precision=2, unit=None):
    """Converts raw bytes to a human friendly format."""
    if power == 1024:
        units = ['B  ', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
    else:
        units = ['B ', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
    if unit in units:
        if power == 1024:
            num = num >> (units.index(unit) * 10)
        else:
            div = power ** units.index(unit)
            num /= float(div)
        return f"{num:.{precision}f}{sep}{unit}"

    for unit in units[:-1]:
        if abs(round(num, precision)) < power:
            return f"{num:.{precision}f}{sep}{unit}"
        num /= float(power)
    return f"{num:.{precision}f}{sep}{units[-1]}"


def to_bytes(size, power=1024, sep=' '):
    """Converts human size to raw bytes."""
    if not size:
        return 0
    size, suffix = size.split(sep, 1)
    suffix = suffix[0]
    try:
        factor = {
            'K': power,
            'M': power**2,
            'G': power**3,
            'T': power**4,
            'P': power**5,
        }[suffix]
    except KeyError:
        factor = 1
    return int(float(size) * factor)


def deep_merge(d, u):
    """Deep merges one dict (u) into another (d)."""
    stack = [(d, u)]
    while stack:
        d, u = stack.pop(0)
        for k, v in u.items():
            if not isinstance(v, collections.abc.Mapping):
                d[k] = v
            else:
                dv = d.setdefault(k, {})
                if not isinstance(dv, collections.abc.Mapping):
                    d[k] = v
                else:
                    stack.append((dv, v))


def jdefault(o):
    return o.__dict__


def debug(func):
    """Print the function signature and return value."""
    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        caller = sys._getframe().f_back.f_code.co_name
        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        logger.debug(f"{caller} called {func.__name__}({signature})")
        value = func(*args, **kwargs)
        logger.debug(f"{func.__name__!r} returned {value!r}")
        return value
    return wrapper_debug



import os
import shutil
import sys
import stat
import time
from PyQt5 import QtWidgets, QtCore
from helper import logger, humansize


class NavCopier(QtWidgets.QWidget):
    """Copy/Move files/directories with a progress window."""
    def __init__(self, arg):
        super().__init__()
        self.act = arg[0]
        self.sources = arg[1: -1]
        self.destination = arg[-1]
        self.copied = 0
        self.total = 0
        self.rate = "0"
        self.total_time = "0 s"
        self.time_elapsed = "0 s"
        self.time_remaining = "0 s"
        self.last_time = 0
        self.build_ui()
        self.copier(self.act)
        sys.exit()

    def build_ui(self):
        """Builds the copy progress window"""
        hbox = QtWidgets.QVBoxLayout()
        self.lbl_action = QtWidgets.QLabel("Checking...")
        self.lbl_src = QtWidgets.QLabel('Source: ' + self.sources[0])
        self.lbl_dest = QtWidgets.QLabel('Destination: ' + self.destination)
        self.pb = QtWidgets.QProgressBar()
        self.lbl_copied = QtWidgets.QLabel(f"Copied Bytes: {self.copied}")
        self.lbl_total = QtWidgets.QLabel(f"Total Bytes: {self.total}")
        self.lbl_remaining = QtWidgets.QLabel(f"Remaining Bytes: {self.total}")
        self.lbl_rate = QtWidgets.QLabel(f"Transfer Rate: {self.rate}")
        self.lbl_time_elapsed = QtWidgets.QLabel(
            f"Time Elapsed: {self.time_elapsed}")
        self.lbl_time_remaining = QtWidgets.QLabel(
            f"Time Remaining: {self.time_remaining}")
        self.pb.setMinimum(0)
        self.pb.setMaximum(100)
        self.pb.setValue(0)
        hbox.addWidget(self.lbl_src)
        hbox.addWidget(self.lbl_dest)
        hbox.addWidget(self.pb)
        hbox.addWidget(self.lbl_copied)
        hbox.addWidget(self.lbl_remaining)
        hbox.addWidget(self.lbl_total)
        hbox.addWidget(self.lbl_rate)
        hbox.addWidget(self.lbl_time_elapsed)
        hbox.addWidget(self.lbl_time_remaining)
        self.setLayout(hbox)
        self.show()

    @QtCore.pyqtSlot()
    def update_progress(self, optional=True):
        """Updates the progress bar."""
        try:
            self.time_elapsed = time.process_time() - self.start_time
            # if optional and self.time_elapsed - self.last_time < 0.1:
            #    return
            self.last_time = time.process_time()
            completed = self.copied / self.total * 100
            self.pb.setValue(completed)
            self.lbl_copied.setText(f"Copied: {humansize(self.copied)}")
            remaining_bytes = self.total - self.copied
            self.lbl_remaining.setText(
                f"Remaining Bytes: {humansize(remaining_bytes)}")
            self.lbl_time_elapsed.setText(
                f"Time Elapsed: {self.time_elapsed}s")
            rate_raw = (self.copied/self.time_elapsed)
            self.rate = '{:.2f} bytes/s'.format(rate_raw)
            self.lbl_rate.setText(f"Transfer rate: {self.rate}")

            time_remaining_raw = remaining_bytes / rate_raw
            self.time_remaining = '{:.2f} s'.format(time_remaining_raw)
            self.lbl_time_remaining.setText(
                f"Time Remaining: {self.time_remaining}")
            QtWidgets.QApplication.processEvents()
        except Exception:
            pass

    def get_size(self, loc, recurse=True):
        """Calculates number of bytes to copied."""
        paste_size = 0
        if os.path.isdir(loc):
            for item in os.scandir(loc):
                paste_size += self.get_size(item, recurse=True)
        else:
            paste_size += os.stat(loc).st_size
        return paste_size

    def copier(self, act):
        """Copy/Move files"""
        self.start_time = time.process_time()
        self.lbl_action.setText("Calculating size...")
        for source in self.sources:
            self.total += self.get_size(source)
        self.lbl_total.setText(f"Total Bytes: {humansize(self.total)}")
        if self.act == "copy":
            for src in self.sources:
                logger.debug(f"Now copying {src}")
                if os.path.isdir(src):
                    self.copytree(src, os.path.join(self.destination,
                                  os.path.basename(src)))
                else:
                    self.lbl_src.setText(f"Source: {src}")
                    self.copy(src, self.destination)
        elif self.act == "move":
            for src in self.sources:
                logger.debug(f"Now moving {src}")
                self.lbl_src.setText(f"Source: {src}")
                self.move(src, self.destination)
        # self.thread_instance.stop()

    def copy(self, src, dst):
        """Reimplemented to report copy progress"""
        if os.path.isdir(dst):
            dst = os.path.join(dst, os.path.basename(src))
        logger.debug(f"{self.act} {src} to {dst}")
        if os.path.exists(dst):
            choice = QtWidgets.QMessageBox.question(
                     None, "File exists",
                     f"{dst} is already present. Overwrite?",
                     QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if choice == QtWidgets.QMessageBox.No:
                return

        self.copyfile(src, dst, follow_symlinks=True)
        shutil.copystat(src, dst, follow_symlinks=True)
        shutil.copymode(src, dst)

    def copyfile(self, src, dst, *, follow_symlinks=True):
        """Reimplemented to report copy progress"""
        if shutil._samefile(src, dst):
            raise shutil.SameFileError("{!r} and {!r} are the same file"
                                       .format(src, dst))

        for fn in [src, dst]:
            try:
                st = os.stat(fn)
            except OSError:
                # File most likely does not exist
                pass
            else:
                # XXX What about other special files? (sockets, devices...)
                if stat.S_ISFIFO(st.st_mode):
                    raise shutil.SpecialFileError("`%s` is a named pipe" % fn)

        if not follow_symlinks and os.path.islink(src):
            os.symlink(os.readlink(src), dst)
        else:
            with open(src, 'rb') as fsrc:
                with open(dst, 'wb') as fdst:
                    self.copyfileobj(fsrc, fdst)
        return dst

    def copyfileobj(self, fsrc, fdst, length=16*1024):
        """Reimplemented to report copy progress"""
        while 1:
            buf = fsrc.read(length)
            if not buf:
                break
            fdst.write(buf)
            self.copied += len(buf)
            self.update_progress()
        self.update_progress(optional=False)

    def copytree(self, src, dst, symlinks=False, ignore=None,
                 copy_function=None, ignore_dangling_symlinks=False):
        """Reimplemented to report copy progress"""
        copy_function = self.copy

        names = os.listdir(src)
        if ignore is not None:
            ignored_names = ignore(src, names)
        else:
            ignored_names = set()

        try:
            os.makedirs(dst)
            logger.debug(f"Directory created: {dst}")
        except FileExistsError:
            choice = QtWidgets.QMessageBox.question(
                        None, "Folder exists",
                        f"{dst} is already present. Merge?",
                        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
            if choice == QtWidgets.QMessageBox.No:
                return
        errors = []
        for name in names:
            if name in ignored_names:
                continue
            srcname = os.path.join(src, name)
            dstname = os.path.join(dst, name)
            try:
                if os.path.islink(srcname):
                    linkto = os.readlink(srcname)
                    if symlinks:
                        # We can't just leave it to `copy_function` because
                        # legacy code with a custom `copy_function` may rely on
                        # copytree doing the right thing.
                        os.symlink(linkto, dstname)
                        shutil.copystat(srcname, dstname,
                                        follow_symlinks=not symlinks)
                    else:
                        # ignore dangling symlink if the flag is on
                        if not os.path.exists(linkto) and \
                                ignore_dangling_symlinks:
                            continue
                        # otherwise let the copy occurs. copy2 will raise error
                        if os.path.isdir(srcname):
                            self.copytree(srcname, dstname, symlinks, ignore)
                        else:
                            copy_function(srcname, dstname)
                elif os.path.isdir(srcname):
                    self.copytree(srcname, dstname, symlinks, ignore,
                                  copy_function)
                else:
                    # Will raise a SpecialFileError for unsupported file types
                    copy_function(srcname, dstname)
            # catch the Error from the recursive copytree so that we can
            # continue with other files
            except shutil.Error as err:
                errors.extend(err.args[0])
            except OSError as why:
                errors.append((srcname, dstname, str(why)))
        try:
            shutil.copystat(src, dst)
        except OSError as why:
            # Copying file access times may fail on Windows
            if getattr(why, 'winerror', None) is None:
                errors.append((src, dst, str(why)))
        if errors:
            raise shutil.Error(errors)
            # logger.debug(errors)
        return dst

    def move(self, src, dst, copy_function=None):
        """Reimplemented to report move progress"""
        copy_function = self.copy
        real_dst = dst
        if os.path.isdir(dst):
            if shutil._samefile(src, dst):
                # We might be on a case insensitive filesystem,
                # perform the rename anyway.
                os.rename(src, dst)
                return

            real_dst = os.path.join(dst, shutil._basename(src))
            if os.path.exists(real_dst):
                if os.path.isdir(real_dst):
                    title = "Folder exists"
                    msg = f"{real_dst} is already present.Merge?"
                else:
                    title = "File exists"
                    msg = f"{real_dst} is already present. Overwrite?"
                choice = QtWidgets.QMessageBox.question(
                        None, title, msg,
                        QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
                if choice == QtWidgets.QMessageBox.No:
                    return
                if os.path.isdir(src):
                    logger.debug(f"now doing a folder move for {src}")
                    for d in os.listdir(src):
                        d2 = os.path.join(src, d)
                        self.move(d2, real_dst)
                else:
                    logger.debug(f"now doing a file move for {src}")
                    # self.move(src, real_dst)
            # raise Error("Destination path '%s' already exists" % real_dst)
        try:
            # if os.path.isdir(real_dst):
            #    raise OSError
            logger.debug(f"Trying rename from {src} to {real_dst}")
            size = self.get_size(src)
            os.rename(src, real_dst)
            self.copied += size
            self.update_progress()  # self.copied)
        except OSError:
            logger.debug(
                f"Rename from {src} to {real_dst} failed. Trying alternatives")
            if os.path.islink(src):
                linkto = os.readlink(src)
                os.symlink(linkto, real_dst)
                os.unlink(src)
            elif os.path.isdir(src):
                if shutil._destinsrc(src, dst):
                    raise shutil.Error(f"Cannot move a directory '{src}' into "
                                       f"itself '{dst}'.")
                try:
                    logger.debug(f"Copy tree from {src} to {real_dst}")
                    self.copytree(src, real_dst, copy_function=copy_function,
                                  symlinks=True)
                    shutil.rmtree(src)
                except Exception:
                    logger.debug(f"Error copying {src}")
            else:
                logger.debug(f"Copy file from {src} to {real_dst}")
                copy_function(src, real_dst)
                os.unlink(src)
        return real_dst


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    ex = NavCopier(sys.argv[1:])
    sys.exit(app.exec_())



#!/bin/python3

import json
import pathlib
import sys
import subprocess
from PyQt5 import QtGui, QtCore, QtWidgets
from .core import Nav, NavView, NavSize
from .custom import NavTree
from .helper import logger, deep_merge
from .navwatcher import NavWatcher
from .panes import NavPane
from .pub import Pub
from .settings import NavSettings


class NavApp(QtWidgets.QApplication):
    """Initialize application."""
    def __init__(self, args):
        super().__init__(args)
        self.window = Navgator(args)
        sys.exit(self.exec_())


class Navgator(QtWidgets.QMainWindow):
    update_sb = QtCore.pyqtSignal(str, int)

    def __init__(self, *args):
        super().__init__()
        self.title = 'Navgator'
        self.load_settings()
        threadpool = QtCore.QThreadPool()
        threadpool.setMaxThreadCount(1)
        try:
            x, y, w, h = Nav.conf["dims"]["main"][:4]
        except Exception:
            screensize = QtWidgets.QDesktopWidget().screenGeometry(-1)
            if "dims" not in Nav.conf:
                Nav.conf["dims"] = {}
            Nav.conf["dims"]["main"] = [0, 0, screensize.width(),
                                        screensize.height()]
            x, y, w, h = Nav.conf["dims"]["main"]
        self.setGeometry(x, y, w, h)
        self.setFocusPolicy(QtCore.Qt.StrongFocus)
        self.define_actions()
        self.sb = self.statusBar()
        self.update_sb.connect(self.status_bar_update)
        self.main_widget = QtWidgets.QWidget(self)
        self.box = QtWidgets.QHBoxLayout(self.main_widget)

        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.splitter1 = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        self.splitter2 = QtWidgets.QSplitter(QtCore.Qt.Vertical)

        self.tree = NavTree()
        self.splitter.insertWidget(0, self.tree)
        self.tree.clicked[QtCore.QModelIndex].connect(self.tree_navigate)

        if not Nav.conf["window"]["main_tree"]:
            self.tree.hide()

        self.panes = []
        self.active_pane = None
        for i in range(1, Nav.conf["panes"]["total"] + 1):
            name = f"Pane {i}"
            try:
                p = NavPane(name, Nav.conf["panes"][name])
            except KeyError:
                Nav.conf["panes"][name] = {
                    'visible': True,
                    'tabs': {
                        'total': 1,
                        '0': {'location': str(pathlib.Path.home())},
                        'active': 0
                    }
                }
                p = NavPane(name, Nav.conf["panes"][name])
            p.pane_updated.connect(self.active_pane_changed)

            if i % 2 != 0:
                self.splitter1.addWidget(p)
            else:
                self.splitter2.addWidget(p)
            if not Nav.conf["panes"][name]["visible"]:
                p.hide()
            self.panes.append(p)

            self.stylesheet = ("QTabBar::tab:selected {{background: {bg};}}"
                               "NavBreadCrumbsBar{{background-color: {bg};}}"
                               "QTabWidget::pane {{border: 0px;}}")
            if Nav.conf["panes"]["active"] == name:
                self.active_pane = p
                p.tabbar.setStyleSheet(self.stylesheet.format(
                    bg=Nav.conf["colors"]["bcbar"]["active"]))
                self.setWindowTitle(f"{self.title} - {name}")
            else:
                p.tabbar.setStyleSheet(self.stylesheet.format(
                    bg=Nav.conf["colors"]["bcbar"]["inactive"]))
        self.splitter.addWidget(self.splitter1)
        self.splitter.addWidget(self.splitter2)
        self.splitter.setSizes(
            Nav.getsizes("tp", [50, 50, 50], [w*0.1, w*0.45, w*0.45]))
        self.splitter1.setSizes(Nav.getsizes("p12", [50, 50], [h*0.5, h*0.5]))
        self.splitter2.setSizes(Nav.getsizes("p23", [50, 50], [h*0.5, h*0.5]))

        # delay creating menubar so that we can hook up to panes and tabs
        self.menubar = self.menuBar()
        self.create_menu()
        self.create_toolbar()
        self.box.addWidget(self.splitter)
        self.setCentralWidget(self.main_widget)
        self.show()
        # self.mainThreadID = QtCore.QThread.currentThread().currentThreadId()
        # Focus active pane and tab
        self.active_pane.tabbar.currentWidget().view.setFocus()
        if Nav.conf["window"]["statusbar"]:
            Pub.subscribe("App", self.update_status_bar)
            self.sb.show()
            self.active_pane_changed(self.active_pane)
            self.sb.showMessage("Ready", 2000)
        else:
            self.sb.hide()

    def define_actions(self):
        self.actions = {
            "exit": {
                "caption": "E&xit",
                "icon": QtGui.QIcon.fromTheme('file-exit'),
                "shortcut": "Ctrl+Q",
                "triggered": self.close,
                "statusTip": "Exit application",
            },
            "back": {
                "caption": "Go &Back",
                "icon": self.style().standardIcon(
                    QtWidgets.QStyle.SP_ArrowBack),
                "shortcut": "Backspace",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              go_back()),
            },
            "forward": {
                "caption": "Go &Forward",
                "icon": self.style().standardIcon(
                    QtWidgets.QStyle.SP_ArrowForward),
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              go_forward()),
            },
            "up": {
                "caption": "Go &Up",
                "icon": self.style().standardIcon(
                    QtWidgets.QStyle.SP_ArrowUp),
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              go_up()),
            },
            "rename": {
                "caption": "&Rename",
                "shortcut": "F2",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().rename_file()),
            },
            "cut": {
                "caption": "C&ut",
                "shortcut": "Ctrl+X",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().copy(cut=True)),
            },
            "copy": {
                "caption": "&Copy",
                "shortcut": "Ctrl+C",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().copy()),
            },
            "paste": {
                "caption": "&Paste",
                "shortcut": "Ctrl+V",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().paste()),
            },
            "trash": {
                "caption": "&Trash",
                "shortcut": "del",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().trash()),
            },
            "delete": {
                "caption": "&Delete",
                "shortcut": "Shift+del",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().delete()),
            },
            "del_dir_up": {
                "caption": "Delete Dir Up",
                "shortcut": "Ctrl+del",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().del_dir_up()),
            },
            "new_file": {
                "caption": "&New File",
                "shortcut": "Ctrl+N",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().new_file("file")),
                "icon": self.style().standardIcon(QtWidgets.QStyle.SP_FileIcon)
            },
            "new_folder": {
                "caption": "New &Folder",
                "shortcut": "Ctrl+Shift+N",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().new_file("folder")),
                "icon": self.style().standardIcon(QtWidgets.QStyle.SP_DirIcon),
            },
            "select_all": {
                "caption": "&Select All",
                "shortcut": "Ctrl+A",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().updateModel(0, 1)),
            },
            "clear_all": {
                "caption": "&Clear All",
                "shortcut": "Ctrl+Shift+A",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().updateModel(0, 0)),
            },
            "invert": {
                "caption": "&Invert Select",
                "shortcut": "Ctrl+Shift+I",
                "triggered": (lambda: self.active_pane.tabbar.
                              currentWidget().invert_selection()),
            },
            "new_tab": {
                "caption": "&New Tab",
                "shortcut": "Ctrl+T",
                "triggered": (lambda: self.active_pane.tabbar.new_tab()),
            },
            "close_tab": {
                "caption": "&Close Tab",
                "shortcut": "Ctrl+W",
                "triggered": (lambda: self.active_pane.tabbar.close_tab()),
            },
            "rename_tab": {
                "caption": "&Rename Tab",
                "shortcut": "Ctrl+E",
                "triggered": (lambda: self.active_pane.tabbar.set_caption(
                              rename=True)),
            },
            "next_tab": {
                "caption": "Ne&xt Tab",
                "shortcut": "Ctrl+Tab",
                "triggered": (lambda: self.active_pane.tabbar.select_tab(
                                      "next")),
            },
            "prev_tab": {
                "caption": "Pre&vious Tab",
                "shortcut": "Ctrl+Shift+Tab",
                "triggered": (lambda: self.active_pane.tabbar.select_tab(
                                      "prev")),
            },
            "close_other_tabs": {
                "caption": "Close &Other Tabs",
                "triggered": (lambda: self.active_pane.tabbar.
                              close_other_tabs()),
            },
            "close_left_tabs": {
                "caption": "Close &Left Tabs",
                "triggered": (lambda: self.active_pane.tabbar.
                              close_left_tabs()),
            },
            "close_right_tabs": {
                "caption": "Close &Right Tabs",
                "triggered": (lambda: self.active_pane.tabbar.
                              close_right_tabs()),
            },
            "maintree": {
                "caption": "&Main Tree",
                "checkable": True,
                "checked": Nav.conf["window"]["main_tree"],
                "shortcut": "F8",
                "triggered": self.tree_toggle,
                "icon": self.style().standardIcon(QtWidgets.QStyle.SP_FileIcon)
            },
            "settings": {
                "caption": "&Settings",
                "shortcut": "F9",
                "triggered": self.show_settings,
            },
            "statusbar": {
                "caption": "&Status Bar",
                "checkable": True,
                "checked": Nav.conf["window"]["statusbar"],
                "triggered": self.statusbar_toggle,
            },
            "copy_path": {
                "caption": "&Copy Path",
                "shortcut": "Ctrl+Shift+P",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              bcbar.copy_path()),
            },
            "paste_and_go": {
                "caption": "&Paste and Go",
                "shortcut": "Ctrl+Shift+G",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              bcbar.paste_and_go()),
            },
            "details_view": {
                "caption": "&Details",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Details))
            },
            "list_view": {
                "caption": "&List",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.List, NavSize.Tiny))
            },
            "small_icons_view": {
                "caption": "&Small Icons",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Icons, NavSize.Small))
            },
            "medium_icons_view": {
                "caption": "&Medium Icons",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Icons, NavSize.Medium))
            },
            "large_icons_view": {
                "caption": "&Large Icons",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Icons, NavSize.Large))
            },
            "xl_icons_view": {
                "caption": "E&xtra Large Icons",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Icons, NavSize.XL))
            },
            "SmallThumbs": {
                "caption": "Small &Thumbnails",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Thumbnails, NavSize.Small))
            },
            "MediumThumbs": {
                "caption": "Medium T&humbnails",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Thumbnails, NavSize.Medium))
            },
            "LargeThumbs": {
                "caption": "Large Th&umbnails",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Thumbnails, NavSize.Large))
            },
            "XLThumbs": {
                "caption": "Extra Large Thum&bnails",
                "triggered": (lambda: self.active_pane.tabbar.currentWidget().
                              switch_view(NavView.Thumbnails, NavSize.XL))
            },
        }

        pane_count = Nav.conf["panes"]["total"]
        if pane_count > 4:
            pane_count = 4
        # items2 = {f"Pane {i}": None for i in range(1, pane_count + 1)}
        for i in range(1, pane_count + 1):
            name = f"Pane {i}"
            self.actions[name] = {
                "caption": f"Pane &{i}",
                "checkable": True,
                "checked": Nav.conf["panes"][name]["visible"],
                "triggered": (lambda a=0, ind=i, name1=name: self.pane_toggle(
                              ind, name1)),
            }
        Nav.actions = {}
        for k, v in self.actions.items():
            cap = v.pop("caption")  # caption isn't valid keyword arg
            Nav.actions[k] = QtWidgets.QAction(cap, self, **v)
            if k in Nav.conf["shortcuts"]:
                Nav.actions[k].setShortcut(Nav.conf["shortcuts"][k])
            self.addAction(Nav.actions[k])

    def create_toolbar(self):
        """Creates a toolbar."""
        self.toolbar = self.addToolBar("Main")
        self.toolbar.addAction(Nav.actions["back"])
        self.toolbar.addAction(Nav.actions["forward"])
        self.toolbar.addAction(Nav.actions["up"])
        for widget in self.toolbar.findChildren(QtWidgets.QWidget):
            widget.installEventFilter(self)

    def create_menu(self):
        """Creates menu recursively."""
        # Python 3.7 tracks insertion order
        items = {
            "file": {"caption": "&File"},
            "edit": {"caption": "&Edit"},
            "view": {"caption": "&View"},
            "tabs": {"caption": "&Tabs"},
            "user": {"caption": "&User", "sm": {}},
            "window": {"caption": "&Window", "sm": {}}
        }

        if "user_commands" not in Nav.conf:
            del items["user"]

        items["file"]["sm"] = [Nav.actions["exit"]]

        items["edit"]["sm"] = [
            Nav.actions["rename"], Nav.actions["cut"], Nav.actions["copy"],
            Nav.actions["paste"], Nav.actions["trash"], Nav.actions["delete"],
            Nav.actions["del_dir_up"], Nav.actions["new_file"],
            Nav.actions["new_folder"],
            {
                "caption": "&Selections",
                "sm": [
                    Nav.actions["select_all"], Nav.actions["clear_all"],
                    Nav.actions["invert"],
                ]
            }
        ]

        items["view"]["sm"] = [
            Nav.actions["details_view"], Nav.actions["list_view"],
            Nav.actions["small_icons_view"], Nav.actions["medium_icons_view"],
            Nav.actions["large_icons_view"], Nav.actions["xl_icons_view"],
            Nav.actions["SmallThumbs"], Nav.actions["MediumThumbs"],
            Nav.actions["LargeThumbs"], Nav.actions["XLThumbs"]
        ]

        items["tabs"]["sm"] = [
            Nav.actions["new_tab"], Nav.actions["close_tab"],
            Nav.actions["rename_tab"],
            Nav.actions["next_tab"], Nav.actions["prev_tab"],
            Nav.actions["close_other_tabs"],
            Nav.actions["close_left_tabs"], Nav.actions["close_right_tabs"],
        ]

        pane_count = Nav.conf["panes"]["total"]
        if pane_count > 4:
            pane_count = 4

        items["window"]["sm"] = [Nav.actions[f"Pane {i}"]
                                 for i in range(1, pane_count + 1)]

        items["window"]["sm"] += [
            Nav.actions["maintree"], Nav.actions["settings"],
            Nav.actions["statusbar"],
        ]

        self.expose_shortcuts(items)
        self.menubar.clear()
        Nav.build_menu(self, items, self.menubar)
        del items

        for m in self.menubar.findChildren(QtWidgets.QMenu):
            if m.title() == "&User":
                self.create_user_menu(m)
                break

    def expose_shortcuts(self, d: dict):
        """Expose shortcuts for customising them."""
        for k, v in d.items():
            if isinstance(v, dict):
                if "shortcut" not in v:
                    self.expose_shortcuts(v)
                else:
                    try:
                        Nav.conf["shortcuts"][k] = v["shortcut"]
                    except KeyError:
                        pass

    def create_user_menu(self, user_menu: QtWidgets.QMainWindow.menuBar):
        """Builds up the user menu based on provided user commands."""
        for k, v in Nav.conf["user_commands"].items():
            user_act = QtWidgets.QAction(k, self)
            if "shortcut" in Nav.conf["user_commands"][k]:
                user_act.setShortcut(Nav.conf["user_commands"][k]["shortcut"])
            command = Nav.conf["user_commands"][k]["command"]
            try:
                args = Nav.conf["user_commands"][k]["args"]
            except (KeyError, TypeError):
                args = None
            user_act.triggered.connect(
                lambda a=0, command=command, args=args:
                self.invoke_user_command(command, args))
            user_menu.addAction(user_act)

    def invoke_user_command(self, command: str, args: str):
        """Invokes customised user commands."""
        if args == "%F":
            item = " ".join(self.active_pane.tabbar.currentWidget().
                            get_selected_items())
        else:
            item = self.active_pane.location
        # logger.debug(f"Command: {command} Args:{item}")
        subprocess.Popen(f'{command} {item}', shell=True)

    def pane_toggle(self, ind: int, item: str):
        """Toggle pane visibility"""
        if Nav.conf["panes"][item]["visible"]:
            # Check if atleast one other visible pane
            for p in self.panes:
                if p != self.panes[ind-1] and p.isVisible():
                    # if hiding active pane, activate another pane
                    if self.panes[ind-1] is self.active_pane:
                        self.active_pane_changed(p)
                    self.panes[ind-1].hide()
                    self.panes[ind-1].set_visibility(False)
                    Nav.conf["panes"][item]["visible"] = False
                    return
            Nav.conf["panes"][item]["visible"] = True
            self.sender().setChecked(True)
            logger.warning(f"Can't hide last visible pane.")
            Pub.notify("App", f"{item}: Can't hide last visible pane.")
            return False
        else:
            self.panes[ind-1].show()
            self.panes[ind-1].set_visibility(True)
            Nav.conf["panes"][item]["visible"] = True

    def tree_toggle(self):
        """Toggle the main tree."""
        if Nav.conf["window"]["main_tree"]:
            self.tree.hide()
        else:
            self.tree.show()
        Nav.conf["window"]["main_tree"] = not Nav.conf["window"]["main_tree"]

    def statusbar_toggle(self):
        """Toggle the status bar."""
        if Nav.conf["window"]["statusbar"]:
            Pub.unsubscribe("App", self.update_status_bar)
            self.sb.hide()
        else:
            Pub.subscribe("App", self.update_status_bar)
            self.sb.show()
        Nav.conf["window"]["statusbar"] = not Nav.conf["window"]["statusbar"]

    def show_settings(self):
        """Displays the settings window."""
        setting = NavSettings(self)
        setting.settings_changed.connect(self.settings_changed)
        setting.show()

    def settings_changed(self):
        logger.debug("Settings were changed")
        self.create_menu()
        # for k in Nav.conf:
        #    if k == "panes":

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.ToolTip:
            if obj.text() == "Go Back":
                obj.setToolTip(self.active_pane.tabbar.currentWidget().
                               history[-1])
            elif obj.text() == "Go Forward":
                try:
                    obj.setToolTip(self.active_pane.tabbar.currentWidget().
                                   tab.future[0])
                except IndexError:
                    obj.setToolTip("No more forward")

            logger.debug(f"caught on {obj} {obj.text()}")
            # return True
        return QtCore.QObject.eventFilter(self, obj, event)

    def closeEvent(self, event):
        """Save and exit application."""
        self.save_settings()
        QtWidgets.QMainWindow.closeEvent(self, event)

    def contextMenuEvent(self, event):
        child = self.childAt(event.pos())
        logger.debug(f"Context menu requested on {child}")

    def save_settings(self):
        """Saves application settings to reload later on."""
        NavWatcher.stop()  # Stop watching folders
        # Remember window sizes
        wind = self.geometry()
        Nav.conf["dims"] = {
            "main": [wind.x(), wind.y(), wind.width(), wind.height()],
            "tp": self.splitter.sizes(),
            "p13": self.splitter1.sizes(),
            "p24": self.splitter2.sizes(),
        }
        # Save pane information
        for p in range(Nav.conf["panes"]["total"]):
            paneid = self.panes[p].pid
            Nav.conf["dims"][paneid] = self.panes[p].splitter.sizes()
            Nav.conf["panes"][paneid]["visible"] = self.panes[p].isVisible()
            Nav.conf["panes"][paneid]["tabs"] = {
                "total": self.panes[p].tabbar.count(),
                "active": self.panes[p].tabbar.currentIndex(), }
            for j in range(Nav.conf["panes"][paneid]["tabs"]["total"]):
                tab = self.panes[p].tabbar.widget(j)
                headers = []
                for h in tab.header:
                    if h.visible:
                        headers.append([h.caption, h.size, h.position])

                Nav.conf["panes"][paneid]["tabs"][j] = {
                        "location": tab.location,
                        "history": list(tab.history),
                        "future": list(tab.future),
                        "caption": self.panes[p].tabbar.tabBar().tabData(j),
                        "sort_column": tab.sort_column,
                        "sort_order": tab.sort_order,
                        "columns": headers,
                        "view": tab.vtype,
                        "itsize": tab.vsize,
                }
            Nav.conf["panes"]["active"] = self.active_pane.pid
        with open(Nav.conf_file, "w") as json_file:
            json.dump(Nav.conf, json_file, indent=4)

    def load_settings(self, conf: str=Nav.conf_file):
        """Load a config file and merge it with defaults"""
        try:
            with open(conf, "r") as json_file:
                d2 = json.load(json_file)
                deep_merge(Nav.conf, d2)
        except (OSError, json.decoder.JSONDecodeError):
            logger.error("{self.title}: Error reading JSON.")

    def tree_navigate(self, index):
        """Navigate to the location selected in the main tree."""
        loc = self.tree.model.filePath(index)
        # logger.debug(f"{self.title}: Tree Navigation: {loc}")
        self.active_pane.navigate(loc)

    @QtCore.pyqtSlot(QtCore.QObject)
    def active_pane_changed(self, obj):
        """Handles pane change event."""
        if self.active_pane.pid != obj.pid:
            self.active_pane.tabbar.setStyleSheet(self.stylesheet.format(
                bg=Nav.conf["colors"]["bcbar"]["inactive"]))
            self.active_pane = obj
            self.setWindowTitle(f"{self.title} - {obj.pid}")
            self.active_pane.tabbar.setStyleSheet(self.stylesheet.format(
                bg=Nav.conf["colors"]["bcbar"]["active"]))
        Nav.actions["up"].setEnabled(self.active_pane.can_go_up)
        Nav.actions["back"].setEnabled(self.active_pane.can_go_back)
        Nav.actions["forward"].setEnabled(self.active_pane.can_go_forward)

    def update_status_bar(self, msg: str, duration: int = 2000):
        """Signals to update main status bar."""
        # Need to emit as signal as this function is invoked from outside
        self.update_sb.emit(msg, duration)

    @QtCore.pyqtSlot(str, int)
    def status_bar_update(self, msg: str, duration: int = 2000):
        """Update main status bar with the provided status information."""
        self.sb.showMessage(msg, duration)

    def keyPressEvent(self, event):
        """Trap key presses to invoke custom methods."""
        # Clear filter on Escape key press
        if event.key() == QtCore.Qt.Key_Escape:
            if self.active_pane.filter_edit.text() != "":
                self.active_pane.filter_edit.setText("")
            else:
                super().keyPressEvent(event)
                event.ignore()
        else:
            super().keyPressEvent(event)
            event.ignore()


def exception_hook(exctype, val, trcbk):
    """Try capturing exceptions and print it to the console."""
    print(exctype, val, trcbk)
    sys._excepthook(exctype, val, trcbk)
    sys.exit(1)


def main(args=None):
    if args is None:
        args = sys.argv[1:]
        logger.debug(args)
    sys._excepthook = sys.excepthook
    NavApp(args)


if __name__ == '__main__':
    main()



import os
import pathlib
from PyQt5 import QtGui, QtCore, QtWidgets
from .core import Nav
from .helper import logger


class NavBreadCrumbMenu(QtWidgets.QLabel):
    """A class to present drop down menu on crumbs down arrow click."""
    open_location = QtCore.pyqtSignal("QString")
    arrow_right_icon = arrow_down_icon = folder_icon = None

    def __init__(self, path):
        super().__init__()
        self.cwd = QtCore.QDir(path)
        if self.arrow_right_icon is None:
            self.__class__.arrow_right_icon = QtGui.QIcon.fromTheme(
                "arrow-right").pixmap(QtCore.QSize(16, 16))
            self.__class__.arrow_down_icon = QtGui.QIcon.fromTheme(
                "arrow-down").pixmap(QtCore.QSize(16, 16))
            self.__class__.folder_icon = QtGui.QIcon.fromTheme("folder")
        self.setPixmap(self.arrow_right_icon)
        self.menu = QtWidgets.QMenu()
        self.menu.setStyleSheet("menu-scrollable: 1;")
        self.menu.aboutToHide.connect(self.onMenuHidden)

    def mousePressEvent(self, event):
        """Populates the drop down menu on arrow button click."""
        self.menu.clear()
        self.cwd.setFilter(QtCore.QDir.NoDotAndDotDot | QtCore.QDir.Dirs)
        if self.cwd.entryList():
            for d in self.cwd.entryList():
                action = self.menu.addAction(self.folder_icon, d)
                action.triggered.connect(self.onMenuItemClicked)
        else:
            action = self.menu.addAction("No folders")
            action.setDisabled(True)

        self.setPixmap(self.arrow_down_icon)
        self.menu.popup(self.mapToGlobal(self.frameRect().bottomLeft()))
        event.accept()

    def onMenuHidden(self):
        """Reverts back the down arrow to right arrow."""
        self.setPixmap(self.arrow_right_icon)

    def onMenuItemClicked(self):
        """Emits a clicked event with location to navigate to."""
        self.open_location.emit(self.cwd.filePath(self.sender().text()))


class NavBreadCrumb(QtWidgets.QLabel):
    """A class to present crumbs on the breadcrumbs bar."""
    open_location = QtCore.pyqtSignal("QString")
    root_icon = None

    def __init__(self, path, current=False):
        super().__init__()
        self.cwd = QtCore.QDir(path)
        if self.root_icon is None:
            self.__class__.root_icon = QtGui.QIcon.fromTheme(
                "drive-harddisk").pixmap(QtCore.QSize(16, 16))

        if self.cwd.isRoot():
            self.setPixmap(self.root_icon)
        else:
            self.setText(self.cwd.dirName())

        if current:
            self.setStyleSheet("QtWidgets.QLabel { font-weight: bold; }")

    def mousePressEvent(self, event):
        """Emits a clicked event with location to navigate to."""
        self.open_location.emit(self.cwd.absolutePath())
        event.accept()


class NavBreadCrumbsBar(QtWidgets.QFrame):
    """A simple class to serve a breadcrumbs bar"""
    clicked = QtCore.pyqtSignal("QString")
    cMenu = None  # common contextMenu across class

    def __init__(self, loc=None):
        super().__init__()
        self.cwd = ""
        lyt = QtWidgets.QHBoxLayout(self)
        lyt.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.setLayout(lyt)
        self.setSizePolicy(QtWidgets.QSizePolicy.Ignored,
                           QtWidgets.QSizePolicy.Fixed)
        if self.cMenu is None:
            self.__class__.cMenu = QtWidgets.QMenu()
            items = [
                Nav.actions["copy_path"],
                Nav.actions["paste_and_go"],
            ]
            Nav.build_menu(self, items, self.cMenu)
        if loc:
            self.create_crumbs(loc)

    def create_crumbs(self, path):
        """Breaks the path into crumbs alongwith dropdown."""
        if not path:
            return
        # Identify common path
        if self.cwd:
            # logger.debug(self.layout().count())
            common = os.path.commonpath([self.cwd, path])
            if common.endswith(os.sep):
                common = common[:-1]
        else:
            common = None
        self.cwd = path
        # logger.debug(f"common: {common}")

        # Remove crumbs which are now invalid. Preserve common ones
        if self.layout().count():
            for i in reversed(range(self.layout().count())):
                try:
                    wid = self.layout().itemAt(i).widget()
                    logger.debug(wid)
                    if common == wid.cwd.absolutePath():
                        break
                    wid.close()
                    wid.deleteLater()
                except AttributeError:
                    self.layout().removeItem(self.layout().itemAt(i))

        if not common:
            count = 0
        else:
            count = len(pathlib.PurePath(common).parts)

        comps = pathlib.PurePath(path).parts
        # Add the uncommon crumbs to serve the breadcrumb bar
        for c in range(count, len(comps)):
            try:
                if not common.endswith(os.sep):
                    common += os.sep
                if comps[c] != os.sep:
                    common += comps[c]
            except Exception:
                continue
            crumb = NavBreadCrumb(common)
            crumb.open_location.connect(self.crumb_clicked)
            menu = NavBreadCrumbMenu(common)
            menu.open_location.connect(self.crumb_clicked)
            self.layout().addWidget(crumb)
            self.layout().addWidget(menu)
        self.layout().addStretch()

    def crumb_clicked(self, loc):
        """Emits a clicked event with location to navigate to."""
        self.clicked.emit(loc)

    def copy_path(self):
        clipboard = QtWidgets.QApplication.clipboard()
        clipboard.setText(self.cwd)

    def paste_and_go(self):
        clip = QtWidgets.QApplication.clipboard().text()
        # if os.path.isdir(clip):
        self.clicked.emit(clip)

    def contextMenuEvent(self, event):
        # child = self.childAt(event.pos())
        self.cMenu.exec_(event.globalPos())



import functools
import operator
# from helper import logger


class Pub:
    sub = {}

    @staticmethod
    def add_subscriber(tree, topic, val):
        """Adds a subscriber to list of subscribers for said topic."""
        k = topic[0]
        if len(topic) == 1:
            tree.setdefault(k, {"sub": []})
            tree[k]["sub"].append(val)
        else:
            if k not in tree:
                tree[k] = {"sub": []}
            Pub.add_subscriber(tree[k], topic[1:], val)
        return tree

    @staticmethod
    def get_subscribers(topics):
        """Gets all subcribers subscribed to the said topic."""
        topics.append("All")
        val = Pub.sub
        out = []
        try:
            for k in topics:
                val = val[k]
                if "sub" in val:
                    out.append(val["sub"])
        except KeyError:
            pass
        if out:
            return functools.reduce(operator.concat, out)
        else:
            return []

    @staticmethod
    def subscribe(topic, fn):
        """Subscribes a callback for the said topic."""
        # logger.debug(f"Subscribe {topic} by {fn}")
        topic = topic.title()
        topics = topic.split(".")
        Pub.sub = Pub.add_subscriber(Pub.sub, topics, fn)

    @staticmethod
    def unsubscribe(topic, fn):
        """Unsubscribes a callback for the said topic."""
        # logger.debug(f"UnSubscribe {topic} by {fn}")
        val = Pub.sub
        topics = topic.split(".")
        for k in topics:
            try:
                val = val[k]
                val["sub"].remove(fn)
            except (KeyError, ValueError):
                pass

    @staticmethod
    def notify(event, *args, **kwargs):
        """Notifies the callbacks when publisher sends a notification."""
        event = event.title()
        # logger.debug(f"Notification received for {event}")
        try:
            for fn in Pub.get_subscribers(event.split(".")):
                try:
                    # logger.debug(f"Notifying for {event} to {fn}")
                    fn(*args, **kwargs)
                except TypeError:
                    pass
        except KeyError:
            pass



from PyQt5 import QtWidgets, QtCore
from .core import Nav, NavStates
from dataclasses import dataclass
# from .helper import logger


@dataclass
class NavColumn:
    caption: str
    size: int
    position: int = -1
    visible: bool = True


class NavTree(QtWidgets.QTreeView):
    """Subclassed for easy instantiating trees across application."""
    def __init__(self):
        super().__init__()
        self.model = QtWidgets.QFileSystemModel(self)
        self.model.setRootPath("/")
        self.model.setResolveSymlinks(False)
        self.model.setFilter(QtCore.QDir.AllDirs | QtCore.QDir.NoDotAndDotDot)
        self.setModel(self.model)
        self.setRootIndex(self.model.index(None))
        self.hideColumn(1)
        self.hideColumn(2)
        self.hideColumn(3)
        self.setHeaderHidden(True)


class NavCheckBoxDelegate(QtWidgets.QStyledItemDelegate):
    """Delegate to handle checkbox toggle"""
    def __init__(self, parent, ind):
        self.parent = parent
        self.ind = ind
        super().__init__(parent)

    def editorEvent(self, event, model, option, index):
        """Check if checkbox was clicked and trigger the toggle."""
        if event.type() == QtCore.QEvent.MouseButtonPress \
                and event.button() == QtCore.Qt.LeftButton:
            offset = self.parent.hv.sectionPosition(self.ind)
            x = event.pos().x()
            if x > offset and x < offset + 20:
                self.parent.view.selectionModel().select(
                    index, QtCore.QItemSelectionModel.Toggle)
                return True
        return False


class NavItemSelectionModel(QtCore.QItemSelectionModel):
    """Subclassed to restrict selection to particular columns."""
    def __init__(self, model, selectable_columns=[0]):
        super().__init__(model)
        self.selectable_columns = selectable_columns

    def select(self, selection, selectionFlags):
        """Reimplemented to prevent selecting some columns."""
        if isinstance(selection, QtCore.QItemSelection):
            indexes = selection.indexes()
            for i in range(len(indexes)):
                index = indexes[i]
                if not index.column() in self.selectable_columns:
                    return
        elif not selection.column() in self.selectable_columns:
            return
        super().select(selection, selectionFlags)


class NavHeaderView(QtWidgets.QHeaderView):
    """Sub-classed to add a checkbox on header."""
    clicked = QtCore.pyqtSignal(int, bool)
    visibility_changed = QtCore.pyqtSignal(int, str, bool)

    def __init__(self, header, orientation=QtCore.Qt.Horizontal, parent=None):
        super().__init__(orientation, parent)
        self.header = header
        self._x_offset = 3
        self._y_offset = 0
        self._width = self._height = 20
        self.cb_cols = 0
        self.isChecked = 0

    def resizeEvent(self, event):
        """Resize table as a whole, required to allow resizing."""
        super().resizeEvent(event)
        self.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        for column in range(0, self.count()):
            self.resizeSection(column, self.header[column].size)
        self.setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.resizeSection(0, self.sectionSize(0))
        self.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
        return

    def paintSection(self, painter, rect, index):
        """Draws a checkbox on the said column in the header."""
        painter.save()
        super().paintSection(painter, rect, index)
        painter.restore()
        self._y_offset = int((rect.height() - self._width) / 2.0)

        if index == self.cb_cols:
            option = QtWidgets.QStyleOptionButton()
            option.rect = QtCore.QRect(rect.x() + self._x_offset,
                                       rect.y() + self._y_offset,
                                       self._width, self._height)
            option.state = QtWidgets.QStyle.State_Enabled | \
                QtWidgets.QStyle.State_Active
            if self.isChecked == 2:
                option.state |= QtWidgets.QStyle.State_NoChange
            elif self.isChecked:
                option.state |= QtWidgets.QStyle.State_On
            else:
                option.state |= QtWidgets.QStyle.State_Off

            self.style().drawControl(QtWidgets.QStyle.CE_CheckBox, option,
                                     painter)

    def updateCheckState(self, state):
        """Updates the check state of the header checkbox."""
        self.isChecked = state
        self.viewport().update()

    def mousePressEvent(self, event):
        """Re-implemented to capture click on the header checkbox ."""
        index = self.logicalIndexAt(event.pos())
        if 0 <= index < self.count():
            x = self.sectionPosition(index) + self._x_offset
            if x < event.pos().x() < x + self._width and \
                    self._y_offset < event.pos().y() < \
                    self._y_offset + self._height:
                if self.isChecked == 1:
                    self.isChecked = 0
                else:
                    self.isChecked = 1
                self.clicked.emit(index, self.isChecked)
                self.viewport().update()
                return
        super().mousePressEvent(event)

    def contextMenuEvent(self, event):
        """Re-implemented to handle the context menu on header."""
        cMenu = QtWidgets.QMenu()
        for h in self.header:
            v = {"checkable": True, "checked": h.visible}
            act = QtWidgets.QAction(h.caption, self, **v)
            cMenu.addAction(act)
        choice = cMenu.exec_(event.globalPos())
        if choice and choice.text() != self.header[0].caption:
            for idx, h in enumerate(self.header):
                if h.caption == choice.text():
                    self.setSectionHidden(idx, h.visible)
                    self.header[idx].visible = not h.visible
                    self.visibility_changed.emit(idx, h.caption, h.visible)
                    return


class NavSortFilterProxyModel(QtCore.QSortFilterProxyModel):
    """Subclassed to provide row numbers and sorting folders to top."""
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.filterString = ''
        self.filterFunctions = {}

    def headerData(self, section, orientation, role):
        """Reimplemented to provide row numbers for vertical headers."""
        # if display role of vertical headers
        if orientation == QtCore.Qt.Vertical and \
                role == QtCore.Qt.DisplayRole:
            return section + 1  # return the actual row number
        # Rely on the base implementation
        return super().headerData(section, orientation, role)

    def lessThan(self, left, right):
        """Reimplemented to sort folders to top."""
        l_data = self.sourceModel().files[left.row()]
        r_data = self.sourceModel().files[right.row()]
        sort_order = self.sortOrder()
        if Nav.conf["sort_folders_first"]:
            l_dir = l_data[self.sourceModel().state] & NavStates.IS_DIR
            r_dir = r_data[self.sourceModel().state] & NavStates.IS_DIR
            try:
                if l_dir > r_dir:
                    return sort_order == QtCore.Qt.AscendingOrder
                elif l_dir < r_dir:
                    return sort_order != QtCore.Qt.AscendingOrder
            except TypeError:
                return True
        try:
            return True if (l_data[left.column()] <= r_data[right.column()]) \
                else False
        except TypeError:
            return True if l_data[left.column()] is None else False
